# -*- coding: utf-8 -*-
#!/usr/bin/env python

import numpy as np
from numpy import power as pow
from scipy.integrate import quad

debug = False


def intinfo(a, mmin, mmax, XMIN, XMAX, xmin, xmax, II, II2):
    ostr = """a={:.2f}, mlim=[{:.2f}, {:.2f}],
Int range=[{:.2f}, {:.2f}],
Requested range=[{:.2f}, {:.2f}],
I0={:.6f}, I1={:.6f}\n""".format(
        a, mmin, mmax, XMIN, XMAX, xmin, xmax, II, II2
    )
    print(ostr.rstrip("\r"))


class Kroupa:
    """An piece-wise powerlaw mass function :math`f(x) \\sim x^{-\\alpha}`
    The PDF in normalized to 1.

    NOTE: only works for len(a) >= 2

    Args:
        a (list): the powerlaw exponents for each piece
        mlim (list): the mass ranges for each piece

    Returns:
        Object

    """

    def __init__(self, a=[1.3, 2.35], mlim=[0.08, 0.5, 120.0]):

        a = np.array(a)
        mlim = np.array(mlim)
        self._a = a
        self._mlim = mlim

        norm = np.zeros(len(a))
        area = np.zeros(len(a))
        C = np.zeros(len(a))

        # Assure piecewise continuity
        C[0] = pow(1.0 / mlim[1], -a[0])  # i=0
        C[1] = pow(1.0 / mlim[1], -a[1])  # i=1
        for i in range(2, len(a)):  # i>1
            C[i] = pow(1.0 / mlim[i], -a[i])
            for j in range(1, i):
                C[i] *= pow((mlim[j + 1] / mlim[j]), -a[j])

        # Loop through pieces to find normalization
        for i in range(len(a)):
            area[i] = self._mom0(mlim[i], mlim[i + 1], a[i])
        norm = area * C
        self._norm = 1.0 / np.sum(norm)
        self._area = norm * self._norm
        self._C = C

    def eval(self, X, N0=1):
        X = np.array([X])  # make it an array
        mlim = self._mlim
        a = self._a
        C = self._C
        norm = self._norm
        Y = np.array([])
        # Loop through pieces, maybe there is a more efficient way, but usually
        # npieces <= 3 for a Kroupa-like IMF so overhead is small
        for i in range(len(a)):
            j = (X >= mlim[i]) * (X < mlim[i + 1])
            y = N0 * norm * C[i] * pow(X[j], -a[i])
            Y = np.r_[Y, y]
        return Y

    def integral(self, xmin, xmax):
        """
        Compute the zeroth and first moment of the PDF
        """

        mlim = self._mlim
        a = self._a
        C = self._C
        norm = self._norm

        if xmin < mlim[0]:
            raise ValueError(
                "xmin is less than the lower domain boundary of {:.2f}".format(mlim[0])
            )
        if xmax > mlim[-1]:
            raise ValueError(
                "xmax is larger than the upper domain boundary of {:.2f}".format(
                    mlim[-1]
                )
            )

        imin = np.where(xmin / mlim >= 1)[0][-1]
        if xmax == mlim[-1]:
            imax = len(mlim) - 1
        else:
            imax = np.where(xmax / mlim < 1)[0][0]

        I = 0
        I2 = 0
        if imin == imax:
            I = norm * C[imin] * self._mom0(xmin, xmax, a[imin])
            I2 = norm * C[imin] * self._mom1(xmin, xmax, a[imin])
            i = imin
            if debug == True:
                print(
                    "part {i} of {j} {imin}, {imax}".format(
                        i=i + 1, j=imax - imin, imin=imin, imax=imax
                    )
                )
                intinfo(a[i], mlim[i], mlim[i + 1], XMIN, XMAX, xmin, xmax, II, II2)
        else:
            for h, i in enumerate(range(imin, imax)):
                XMIN = np.max((mlim[i], xmin))
                XMAX = np.min((mlim[i + 1], xmax))
                II = norm * C[i] * self._mom0(XMIN, XMAX, a[i])
                II2 = norm * C[i] * self._mom1(XMIN, XMAX, a[i])
                I += II
                I2 += II2
                if debug == True:
                    print(
                        "part {h} of {j} {imin}, {imax}".format(
                            h=h + 1, j=imax - imin, imin=imin, imax=imax
                        )
                    )
                    intinfo(a[i], mlim[i], mlim[i + 1], XMIN, XMAX, xmin, xmax, II, II2)
        return (I, I2)

    def sample(self, n):
        """Sample n stars from the mass function.

        Args:
            n (int): number of points to be sampled
        Return:
            mass (float or array): sampled mass values.
        """
        mlim = self._mlim
        a = self._a
        C = self._C
        norm = self._norm
        area = self._area

        X = np.random.rand(n)
        mass = np.array([])
        for i in range(len(a)):
            j = (X >= np.sum(area[0:i])) * (X < np.sum(area[0 : i + 1]))
            x = np.random.rand(len(X[j]))
            m = self._getmass(x, a[i], mlim[i], mlim[i + 1])
            mass = np.r_[mass, m]
        return mass

    def _mom0(self, xmin, xmax, a):
        """ First moment """
        if a == 1:
            return np.log(xmin) - np.log(xmax)
        else:
            return (pow(xmax, 1.0 - a) - pow(xmin, 1.0 - a)) / (1.0 - a)

    def _mom1(self, xmin, xmax, a):
        """ Second moment """
        if a == 0:
            return np.log(xmin) - np.log(xmax)
        else:
            return (pow(xmax, 2.0 - a) - pow(xmin, 2.0 - a)) / (2.0 - a)

    def _getmass(self, x, slope, xmin, xmax):
        """
        This is the sampler for a single power-law. In the form
        :math:`\\xi(m) = A\\,m^{\\alpha}`.

        Args:
            x (float or array): random variable
            slope (float): :math:`\\alpha` slope of the power law
            xmin, xmax (float, float): domain of the power-law PDF
        Return:
            mass (float or array): sampled mass values.
        """
        if slope == 1:
            A = np.log(xmax) - np.log(xmin)
        else:
            A = (1.0 / (1 - slope)) * (pow(xmax, 1.0 - slope) - pow(xmin, 1 - slope))
        mass = (1.0 - slope) * x * A + xmin ** (1.0 - slope)
        Z = 1.0 / (1.0 - slope)
        mass = mass ** Z
        return mass


if __name__ == "__main__":

    from scipy.integrate import quad, quadrature
    from scipy.interpolate import interp1d
    import matplotlib.pyplot as p
    from sys import argv

    a = np.r_[np.zeros(5) + 1.3, np.zeros(10) + 2.3]
    mlim = np.r_[
        np.logspace(np.log10(0.08), np.log10(0.5), 6),
        np.logspace(np.log10(0.5), np.log10(120), 11)[1:],
    ]

    K = Kroupa(a=a, mlim=mlim)
    mlim = K._mlim

    eep, m, logT, logL, g, r, i = np.loadtxt(
        "/home/balbinot/ssptools/ssptools/data/models/MIST_iso_5b0533fab00b3.iso.cmd",
        usecols=(0, 2, 4, 6, 10, 11, 12),
        unpack=True,
    )

    ilogL = interp1d(m, logL, bounds_error=False)
    ilogT = interp1d(m, logT, bounds_error=False)
    ig = interp1d(m, g, bounds_error=False)
    ir = interp1d(m, r, bounds_error=False)
    ii = interp1d(m, i, bounds_error=False)

    def evalf(m):
        return K.eval(m) * (np.exp(ilogL(m)))

    # Test integration
    for mmin, mmax in zip(m[0:-1], m[1:]):
        I, I2 = K.integral(mmin, mmax)
        print(
            "{:.8f} {:.8f} {:.8f} {:.8f} {:.8f}".format(mmin, mmax, mmax - mmin, I, I2)
        )

    # Show synthetic CMD
    N = np.int(3.0e5)
    p.figure(figsize=(8, 8))
    mass = K.sample(N)  # sample some masses

    sg = ig(mass)
    sr = ir(mass)

    def err(x):
        ee = 0.002 + np.exp((x - 12) / 1.3)
        ee[ee > 0.5] = 0.5
        return ee

    sg += err(sg) * np.random.randn(len(sg))
    sr += err(sr) * np.random.randn(len(sr))

    p.plot(sg - sr, sg, "k.", ms=1, alpha=0.8)
    p.ylim(14, -2)
    p.xlim(-1.2, 3)

    # Show MF
    p.figure(figsize=(8, 8))
    X = np.linspace(mlim[0], mlim[-1] - 0.01, 10000)
    Y = K.eval(X, N0=N) / N
    print("average mass {:.4f}".format(np.mean(mass)))
    bins = np.linspace(np.log10(mlim[0]), np.log10(mlim[-1]), 100)
    p.hist(
        np.log10(mass),
        bins=bins,
        normed=True,
        log=True,
        histtype="stepfilled",
        color="red",
        alpha=0.5,
    )

    p.plot(np.log10(X), X * np.log(10) * Y, "k-")
    p.xlabel(r"log(m/M$_{\odot})$")
    p.ylabel(r"log($\xi$(m))")
    p.show()



#!/usr/bin/env python
#-*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as p
from scipy.interpolate import LinearNDInterpolator as interpnd

models = 'stdHe_afe2.all'

a,Z,Zeff,feh,afe,Y,m,logg,V,I,eep = np.loadtxt(models,
                                               usecols=(0,1,2,3,4,5,7,9,16,21,6),
                                               unpack=True)

feh = np.round(feh,1) ## To make life easier.
FEHS = np.unique(feh)
p.subplot(121)
for A in [7.0]:
    for B in np.unique(feh):
        print A, B
        jj = (feh==B)*(a==A)*(m>0.3)
        p.plot(V[jj]- I[jj], I[jj], '-', lw=2)
p.xlabel('V-I (fixed age)')
p.ylabel('V')
p.ylim(p.ylim()[::-1])

p.subplot(122)
for A in np.unique(a):
    for B in [FEHS[3]]:
        print A, B, 'fixfeh'
        jj = (feh==B)*(a==A)*(m>0.3)
        p.plot(V[jj]- I[jj], I[jj], '-', lw=2)
p.xlabel('V-I (fixed [Fe/H])')
p.ylabel('V')
p.ylim(p.ylim()[::-1])
p.show()


## Trim models for speed
j = (feh<=0)*(m>0.3)
N = len(a[j])
pts = np.zeros((N,3))
pts[:,0] = eep[j]
pts[:,1] = a[j]
pts[:,2] = feh[j]

jj = (a==5)*(feh==-1.5)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 5 Gyr, feh=-1.5')
jj = (a==5)*(feh==-2.5)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 5 Gyr, feh=-2.5')
jj = (a==5)*(feh==-2.0)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 5 Gyr, feh=-2.0')

jj = (a==12)*(feh==-2.0)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 12 Gyr, feh=-2.0')
jj = (a==12)*(feh==-2.5)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 12 Gyr, feh=-2.5')
jj = (a==12)*(feh==-1.5)
p.plot(V[jj]- I[jj], I[jj], '-', lw=2, label='age = 12 Gyr, feh=-1.5')


#tri = np.load('asd.tri.npy')

iV = interpnd(pts, V[j])
iI = interpnd(iV.tri, I[j])

np.save('asd.tri', iV.tri)

FEH = -2.5
AGE = 5.0
ms = np.arange(1,300, 0.1) ## actually eep
for feh in [-1.75, -2.25]:
    p.plot(iV(ms, AGE,feh)-iI(ms, AGE,feh), iI(ms, AGE,feh), 'k:', label='age = {:.2f} Gyr, feh={:.2f}'.format(AGE, feh))
FEH = -2.5
AGE = 12.0
ms = np.arange(1,300, 0.1) ## actually eep
for feh in [-1.75, -2.25]:
    p.plot(iV(ms, AGE,feh)-iI(ms, AGE,feh), iI(ms, AGE,feh), 'k:', label='age = {:.2f} Gyr, feh={:.2f}'.format(AGE, feh))

p.xlabel('V-I')
p.ylabel('V')
p.legend(loc='best')
p.ylim(p.ylim()[::-1])
p.show()
exit()




#!/usr/bin/env python

import pytest
import numpy as np

from ssptools import ifmr


metals = [-3.00, -2.50, -2.00, -1.00, -0.50, 0.50, 1.00]


class TestMetallicity:
    '''Tests about IFMR metallicity and bound checks'''

    WD_metals = [-2.00, -2.00, -2.00, -1.00, -0.50, -0.50, -0.50]
    BH_metals = [-2.50, -2.50, -2.00, -1.00, -0.50, 0.50, 0.50]

    @pytest.mark.parametrize('FeH', metals)
    def test_stored_FeH(self, FeH):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.FeH == FeH

    @pytest.mark.parametrize("FeH, expected", zip(metals, WD_metals))
    def test_WD_FeH(self, FeH, expected):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.FeH_WD == expected

    @pytest.mark.parametrize("FeH, expected", zip(metals, BH_metals))
    def test_BH_FeH(self, FeH, expected):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.FeH_BH == expected


class TestBounds:
    '''Tests about IFMR remnant mass bounds'''

    # Initial and final mass bounds for all remnants, for each FeH in `metals`
    BH_mi = [(19.7, 150.0), (19.7, 150.0), (19.8, 150.0), (20.5, 150.0),
             (20.9, 150.0), (24.3, 150.0), (24.3, 150.0)]

    BH_mf = [(5.4977, np.inf), (5.4977, np.inf), (5.5952, np.inf),
             (5.503, np.inf), (5.5386, np.inf), (5.5081, np.inf),
             (5.5081, np.inf)]

    WD_mi = [(0.0, 5.318525), (0.0, 5.318525), (0.0, 5.318525),
             (0.0, 5.47216), (0.0, 5.941481), (0.0, 5.941481), (0.0, 5.941481)]

    WD_mf = [(0.0, 1.228837), (0.0, 1.228837), (0.0, 1.228837),
             (0.0, 1.228496), (0.0, 1.256412), (0.0, 1.256412), (0.0, 1.256412)]

    NS_mi = [(5.318525, 19.7), (5.318525, 19.7), (5.318525, 19.8),
             (5.472164, 20.5), (5.941481, 20.9), (5.941481, 24.3),
             (5.941481, 24.3)]

    @pytest.mark.parametrize("FeH, mi", zip(metals, BH_mi))
    def test_BH_mi(self, FeH, mi):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.BH_mi == pytest.approx(mi)

    @pytest.mark.parametrize("FeH, mf", zip(metals, BH_mf))
    def test_BH_mf(self, FeH, mf):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.BH_mf == pytest.approx(mf)

    @pytest.mark.parametrize("FeH, mi", zip(metals, WD_mi))
    def test_WD_mi(self, FeH, mi):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.WD_mi == pytest.approx(mi)

    @pytest.mark.parametrize("FeH, mf", zip(metals, WD_mf))
    def test_WD_mf(self, FeH, mf):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.WD_mf == pytest.approx(mf)

    @pytest.mark.parametrize("FeH, mi", zip(metals, NS_mi))
    def test_NS_mi(self, FeH, mi):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.NS_mi == pytest.approx(mi)

    @pytest.mark.parametrize("FeH", metals)
    def test_NS_mf(self, FeH):
        IFMR = ifmr.IFMR(FeH)
        assert IFMR.NS_mf == pytest.approx((1.4, 1.4))


class TestPredictions:
    '''Tests about IFMR remnant mass and type predictions'''

    IFMR = ifmr.IFMR(FeH=-1.00)

    ini_masses = np.geomspace(0.2, 100, 25)

    rem_masses = np.array([
        0.296115, 0.403242, 0.485982, 0.536884, 0.560416,  # White Dwarfs
        0.576329, 0.608092, 0.654573, 0.686416, 0.731103,
        0.897034, 1.05756359, 1.10926924,
        1.4, 1.4, 1.4, 1.4, 1.4,  # Neutron Stars
        17.413826, 10.313890, 19.816023, 15.147086, 21.023453,  # Black Holes
        29.188165, 40.681
    ])

    rem_types = (['WD'] * 13) + (['NS'] * 5) + (['BH'] * 7)

    def test_predict_mass(self):
        mf = self.IFMR.predict(self.ini_masses)
        assert mf == pytest.approx(self.rem_masses, abs=1e-5)

    def test_predict_type(self):
        mt = self.IFMR.predict_type(self.ini_masses)
        assert mt == self.rem_types



#!/usr/bin/env python

import warnings

import pytest
import numpy as np

from ssptools import evolve_mf, masses, ifmr


# Mixture of `ssptools` and `GCfit` defaults for `evolve_mf`
DEFAULT_KWARGS = dict(
    m_breaks=[0.1, 0.5, 1.0, 100], a_slopes=[-0.5, -1.3, -2.5],
    nbins=[5, 5, 20], FeH=-1.00, tout=[12_000], Ndot=0.,
    N0=5e5, tcc=0.0, NS_ret=0.1, BH_ret_int=1.0, BH_ret_dyn=1.0,
    natal_kicks=False, vesc=90
)


class TestHelperMethods:

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        emf = evolve_mf.EvolvedMF(**DEFAULT_KWARGS)

    # ----------------------------------------------------------------------
    # Testing computation of t_ms lifetime
    # ----------------------------------------------------------------------

    def test_tms_values(self):
        mi = [0.5, 1.0, 100.0]

        tms = self.emf.compute_tms(mi)

        expected = np.array([8.10364433e+04, 5.72063993e+03, 1.80376582e+00])

        assert tms == pytest.approx(expected)

    @pytest.mark.filterwarnings("ignore:.*:RuntimeWarning")
    @pytest.mark.parametrize(
        'mi, expected',
        [
            (-1.0, np.nan),
            (0, np.inf),
            (1e-15, np.inf),
            (np.finfo('float64').max, 0.18799593)  # FeH=-1 asympt. at ~0.187995
        ],
        ids=['negative', 'zero', 'near-zero', 'near-inf']
    )
    def test_tms_bounds(self, mi, expected):

        tms = self.emf.compute_tms(mi)

        assert tms == pytest.approx(expected, nan_ok=True)

    def test_tms_sort(self):
        mi = np.sort(np.random.random(100) * 100)

        tms = self.emf.compute_tms(mi)

        assert np.all(tms[:-1] >= tms[1:])

    # ----------------------------------------------------------------------
    # Testing computation of m_to turnoff mass
    # ----------------------------------------------------------------------
    # mto function asymptotes at a0, then is imaginary to the left, except at 0

    def test_mto_values(self):
        ti = np.array([8.10364433e+04, 5.72063993e+03, 1.80376582e+00])

        mto = self.emf.compute_mto(ti)

        expected = [0.5, 1.0, 100.0]

        assert mto == pytest.approx(expected)

    @pytest.mark.filterwarnings("ignore:.*:RuntimeWarning")
    @pytest.mark.parametrize(
        'ti, expected',
        [
            (-1.0, np.inf),
            (0, np.inf),
            (emf._tms_constants[0] - 1e-15, np.inf),
            (emf._tms_constants[0], np.inf),
            (np.finfo('float64').max, 0.0)
        ],
        ids=['negative', 't = 0', 't < a0', 't = a0', 'near-inf']
    )
    def test_mto_bounds(self, ti, expected):

        mto = self.emf.compute_mto(ti)

        assert mto == pytest.approx(expected, nan_ok=True)

    def test_mto_sort(self):
        ti = np.sort(np.random.random(100) * 100) + self.emf._tms_constants[0]

        mto = self.emf.compute_mto(ti)

        assert np.all(mto[:-1] >= mto[1:])

    def test_mto_tms_inverse(self):
        mi = [0.5, 1.0, 100.0]

        mf = self.emf.compute_mto(self.emf.compute_tms(mi))

        assert mf == pytest.approx(mi)

    # ----------------------------------------------------------------------
    # Testing computation of P_k helper integral solution
    # ----------------------------------------------------------------------

    # @pytest.mark.parametrize('k', [1., 1.5, 2.])
    # @pytest.mark.parametrize('a', [-2, -1., -0.5, 1.0])
    # def test_Pk(self, a, k):
    #     from scipy.integrate import quad

    #     m1, m2 = 0.5, 1.0
    #     expected, err = quad(lambda m: m**(a + k - 1), m1, m2)

    #     Pk = self.emf._Pk(a=a, k=k, m1=m1, m2=m2)

    #     assert Pk == pytest.approx(expected, abs=err)


class TestBHEvolution:

    emf_kw = DEFAULT_KWARGS.copy() | {'tout': [0.]}

    @pytest.fixture()
    def Mi(self):
        return np.array([10., 10., 10.])

    @pytest.fixture()
    def Ni(self):
        return np.array([20., 10., 5.])

    # ----------------------------------------------------------------------
    # Test BH natal kick routines
    # ----------------------------------------------------------------------

    @pytest.mark.parametrize(
        'vesc, expected',
        [
            (25., np.stack((np.array([0.002227, 0.002227, 0.002779]),
                            np.array([0.004454, 0.002227, 0.001389])))),
            (100., np.stack((np.array([0.136963, 0.136963, 0.169875]),
                             np.array([0.273926, 0.136963, 0.084937])))),
            (200., np.stack((np.array([0.966443, 0.966443, 1.175480]),
                             np.array([1.932887, 0.966443, 0.587740])))),
        ],
        ids=[f'vesc = {v}' for v in (25, 100, 200)]
    )
    def test_natal_kick_quantities(self, Mi, Ni, vesc, expected):

        # Results not actually reliant on this evolution, just need the object
        kw = self.emf_kw | {'vesc': vesc, 'natal_kicks': True}
        emf = evolve_mf.EvolvedMF(**kw)

        Mf, Nf, _ = emf._natal_kick_BH(Mi, Ni)

        assert np.stack((Mf, Nf)) == pytest.approx(expected, rel=1e-3)

    @pytest.mark.parametrize(
        'vesc, expected',
        [
            (25., 29.992765),
            (100., 29.556198),
            (200., 26.891631),
        ],
        ids=[f'vesc = {v}' for v in (25, 100, 200)]
    )
    def test_natal_kick_total(self, Mi, Ni, vesc, expected):

        kw = self.emf_kw | {'vesc': vesc, 'natal_kicks': True}
        emf = evolve_mf.EvolvedMF(**kw)

        _, _, ejected = emf._natal_kick_BH(Mi, Ni)

        assert ejected == pytest.approx(expected, rel=1e-3)

    # ----------------------------------------------------------------------
    # Test BH dynamical ejection routines
    # ----------------------------------------------------------------------

    base_emf = evolve_mf.EvolvedMF(**emf_kw)

    @pytest.mark.parametrize(
        'M_eject, expected',
        [
            (0., np.stack((np.array([10., 10, 10]), np.array([20., 10, 5])))),
            (15., np.stack((np.array([10., 5, 0]), np.array([20., 5, 0])))),
            (30., np.stack((np.array([0., 0, 0]), np.array([0., 0, 0])))),
        ],
    )
    def test_dyn_ej_quantities(self, Mi, Ni, M_eject, expected):

        Mf, Nf = self.base_emf._dyn_eject_BH(Mi, Ni, M_eject=M_eject)

        assert np.stack((Mf, Nf)) == pytest.approx(expected)

    def test_dyn_ej_overflow(self, Mi, Ni):

        M_eject_overflow = Mi.sum() + 0.1

        with pytest.raises(ValueError):
            self.base_emf._dyn_eject_BH(Mi, Ni, M_eject=M_eject_overflow)


class TestfBHInit:

    @pytest.fixture()
    def Mtot(self):
        return 100.

    @pytest.fixture()
    def Mi(self):
        return np.array([10., 10., 10.])

    @pytest.fixture()
    def Ni(self):
        return np.array([20., 10., 5.])

    # ----------------------------------------------------------------------
    # Test target f_BH version of BH dynamical ejection routines
    # ----------------------------------------------------------------------

    emf_kw = DEFAULT_KWARGS.copy() | {'tout': [100.]}

    @pytest.mark.parametrize(
        'fBH_target, expected',
        [
            (1, np.stack((np.array([10, 10, 10.]), np.array([20, 10, 5.])))),
            (0.2, np.stack((np.array([10, 7.5, 0.]), np.array([20, 7.5, 0.])))),
            (0.125, np.stack((np.array([10, 0., 0.]), np.array([20, 0., 0.])))),
            (0.0, np.stack((np.array([0., 0., 0.]), np.array([0., 0., 0.])))),
        ],
    )
    def test_dyn_ej_quantities(self, Mi, Ni, Mtot, fBH_target, expected):

        emf = evolve_mf.EvolvedMFWithBH(f_BH=0, **self.emf_kw)

        Mf, Nf = emf._dyn_eject_BH(Mi, Ni, Mtot=Mtot, fBH_target=fBH_target)

        assert np.stack((Mf, Nf)) == pytest.approx(expected)

    # ----------------------------------------------------------------------
    # Test f_BH init and final mass fraction
    # ----------------------------------------------------------------------

    def test_fBH_overflow(self):

        f_BH = 1.1

        with pytest.raises(ValueError):
            evolve_mf.EvolvedMFWithBH(f_BH=f_BH, **self.emf_kw)

    def test_fBH_underflow(self):

        f_BH = -0.1

        with pytest.raises(ValueError):
            evolve_mf.EvolvedMFWithBH(f_BH=f_BH, **self.emf_kw)

    @pytest.mark.parametrize('fBH_target', [0.01, 0.0])
    def test_final_fBH_full(self, fBH_target):

        mf = evolve_mf.EvolvedMFWithBH(f_BH=fBH_target, **self.emf_kw)

        mf.Nmin = 0.0  # avoid those pesky small-N bins

        final_fBH = mf.M[mf.types == 'BH'].sum() / (mf.M).sum()

        assert fBH_target == pytest.approx(final_fBH)

    @pytest.mark.parametrize('fBH_target', [0.01, 0.0])
    def test_final_fBH(self, fBH_target):

        mf = evolve_mf.EvolvedMFWithBH(f_BH=fBH_target, **self.emf_kw)

        final_fBH = mf.M[mf.types == 'BH'].sum() / (mf.M).sum()

        # Even though it's not perfect, don't let it get too far off.
        assert fBH_target == pytest.approx(final_fBH, rel=0.001)


# class TestMassEvolution:
class TestDerivatives:

    # TODO tests for the _set_imf and _evolve functions, the only ones which
    #   modify in-place, so I haven't touched them here

    emf_kw = DEFAULT_KWARGS.copy() | {'tout': [0.], 'nbins': [1, 1, 2]}

    mb = masses.MassBins(emf_kw['m_breaks'], emf_kw['a_slopes'],
                         emf_kw['nbins'], emf_kw['N0'],
                         ifmr.IFMR(emf_kw['FeH']))

    # ----------------------------------------------------------------------
    # Derivative routines
    # ----------------------------------------------------------------------

    @pytest.fixture()
    def y(self):
        nb = self.mb.nbin
        return np.array(([1000] * nb.MS) + ([1] * nb.MS)
                        + ([500] * nb.WD) + ([10] * nb.NS) + ([100] * nb.BH)
                        + ([250] * nb.WD) + ([140] * nb.NS) + ([250] * nb.BH))

    @pytest.mark.parametrize(
        't, expected',
        [
            (100, np.array([0., 0., -10.16057518, 0.,
                            0., 0., 0., 0.,
                            0., 0., 10.16057518, 0.,
                            0., 0., 0., 0.,
                            11.28587621, 0., 0., 0.])),
            (12000, np.array([0., -0.07375997, 0., 0.,
                              0., 0., 0., 0.,
                              0., 0.07375997, 0., 0.,
                              0., 0., 0., 0.04325484,
                              0., 0., 0., 0.])),
        ],
    )
    def test_sev(self, y, t, expected):

        emf = evolve_mf.EvolvedMF(**self.emf_kw)

        ydot = emf._derivs_sev(t, y)
        assert ydot == pytest.approx(expected)

    @pytest.mark.parametrize(
        't, expected',
        [
            (100, np.array([-3.92479847e+00, -1.65079648e+00, 0.00000000e+00,
                            0.00000000e+00, 1.84468190e-03, 3.20949877e-03,
                            0.00000000e+00, 0.00000000e+00, -1.47480168e+00,
                            -1.47480168e+00, -1.47480168e+00, 0.00000000e+00,
                            0.00000000e+00, 0.00000000e+00, -7.37400842e-01,
                            -7.37400842e-01, -7.37400842e-01, 0.00000000e+00,
                            0.00000000e+00, 0.00000000e+00])),
            (12000, np.array([-3.74567798e+00, -2.03183858e+00, 0.00000000e+00,
                              0.00000000e+00, 1.76049405e-03, 2.89876603e-03,
                              0.00000000e+00, 0.00000000e+00, -1.40749448e+00,
                              -1.40749448e+00, -1.40749448e+00, 0.00000000e+00,
                              0.00000000e+00, 0.00000000e+00, -7.03747241e-01,
                              -7.03747241e-01, -7.03747241e-01, 0.00000000e+00,
                              0.00000000e+00, 0.00000000e+00])),
        ],
    )
    def test_esc(self, y, t, expected):

        kw = self.emf_kw | {'Ndot': -10}
        emf = evolve_mf.EvolvedMF(**kw)

        ydot = emf._derivs_esc(t, y)
        assert ydot == pytest.approx(expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import collections.abc
from collections import namedtuple

import numpy as np


mbin = namedtuple('mbin', ('lower', 'upper'))
rem_classes = namedtuple('rem_classes', ('WD', 'NS', 'BH'))
star_classes = namedtuple('star_classes', ('MS',) + rem_classes._fields)


@np.errstate(invalid='ignore')
def Pk(a, k, m1, m2):
    r'''Convenience function for computing quantities related to IMF

    ..math ::
        \begin{align}
            P_k(\alpha_j,\ m_{j,1},\ m_{j,2})
                &= \int_{m_{j,1}}^{m_{j,2}} m^{\alpha_j + k - 1}
                    \ \mathrm{d}m \\
                &= \begin{cases}
                    \frac{m_{j,2}^{\alpha_j+k} - m_{j,1}^{\alpha_j+k} }
                         {\alpha_j + k},
                         \quad \alpha_j + k \neq 0 \\
                    \ln{\left(\frac{m_{j,2}}{m_{j,1}}\right)},
                        \quad \alpha_j + k = 0
                \end{cases}
        \end{align}

    Parameters
    ----------
    a : float
        Mass function power law slope effective between m1 and m2

    k : int
        k-index

    m1, m2 : float
        Upper and lower bound of given mass bin or range
    '''

    a = np.asarray(a, dtype=float)
    res = np.asarray((m2 ** (a + k) - m1 ** (a + k)) / (a + k))  # a != k

    if (casemask := np.asarray(-a == k)).any():
        res[casemask] = np.log(m2 / m1)[casemask]

    return res


def _divide_bin_sizes(N, Nsec):
    '''Split N into Nsec as equally as possible'''
    Neach, ext = divmod(N, Nsec)
    return ext * [Neach + 1] + (Nsec - ext) * [Neach]


class MassBins:
    '''Representation and handling of mass bins, for both stars and remnants

    A class handling the construction, holding and handling of mass bins.
    Both stars and remnants mass bins are handled separately.

    Based on input parameters defining the binned mass function (break masses,
    number of bins, etc.) sets up the spacing and boundaries of mass bins.
    Mass bins are defined using the `mbin` named tuple, with upper and lower
    bounds for each bin.

    The spacing of the bins, currently, can be either linear or log spaced,
    between the break masses.

    Remnant bins are defined based on the possible remnant max/min bounds,
    dictated by the IFMR, and the stellar mass bins.
    Remnant bins are stored separately for each remnant class, avoiding any
    possible overlap between different kinds of remnants.

    Methods are provided to help with interacting with these mass bins,
    including separating the different types, determining which bin a mass falls
    into, etc.

    This class provides a number of methods to help and facilitate the use of
    numerical ODE solvers, such as `scipy.integrate.ode` in `EvolvedMF`, with
    the mass bins.
    Such solvers are based on a single array of `y` numerical values (such that
    :math:$y'(t)=f(t,y)$), which must be dissected within the derivative
    functions into their constituent values.
    In the case of the derivatives in `EvolvedMF`, this `y` array is made up of
    8 parameters (arrays); Ns, the number of stars in each stellar mass bin,
    alphas, the mass function slopes in each stellar mass bin, Nwd, Nns, Nbh,
    the number of remnants in each remnant mass bin, Mwd, Mns, Mbh, the total
    mass of remnants in each remnant mass bin. Each of these constituent arrays
    have the size of the corresponding number of mass bins.
    Methods are available here to pack these distinct arrays into a single `y`
    or unpack a single `y` into the separate arrays.

    Note: currently this class only supports an IMF of a 3-component power
    law.

    Parameters
    ----------
    m_break : list of float
        IMF break-masses (including outer bounds; size 4)

    a : list of float
        IMF power law slopes (size 3)

    nbins : int or list of int (size 3) or dict
        Number of stellar mass bins in each regime. If a single integer, the
        number mass bins between each break mass will be equal, otherwise
        the number of bins between each can be specified directly.
        Remnant bins will be created between IFMR bounds by copying bins from
        the created stellar mass bins.
        Also allowed is a dict of {'MS', 'WD', 'NS', 'BH'}, each containing
        a number of bins, where the remnant bins can be specified directly.

    N0 : int
        Total initial number of stars

    ifmr : ifmr.IFMR
        Initial-final mass relation class, required to set remnant mass
        boundaries

    binning_method : {'default', 'split_log', 'split_linear'}, optional
        The spacing method to use when constructing the mass bins.
        The default method ('split_log') will logspace the mass bins between
        each break mass, restarting the spacing in each regime.
        'split_linear' will linearly space the bins between the break masses.
        Note that in both cases the spacing between the regimes will *not* be
        the same

    Attributes
    ----------
    nbin : star_classes of int
        The number of bins in each stellar object type

    nbin_tot : int
        The total number of bins, including stars and remnants

    bins : star_classes of mbin of ndarray
        Mass bins of each stellar object type, defined using the `mbin` named
        tuple class, with upper and lower bounds. Each will have a size
        corresponding to `nbin`.

    '''

    def __init__(self, m_break, a, nbins, N0, ifmr, *,
                 binning_method='default'):

        N_MS_breaks = len(m_break) - 1

        self.a, self.m_break, self.N0 = a, m_break, N0

        # ------------------------------------------------------------------
        # Unpack number of stellar bins
        # ------------------------------------------------------------------

        # Bins are specified in dict for each type
        if isinstance(nbins, collections.abc.Mapping):
            try:
                nbin_MS = nbins['MS']

            except KeyError as err:
                mssg = f"Missing required stellar type in `nbins`: {err}"
                raise KeyError(mssg)

        # Only the stellar bins are specified, implicitly
        else:
            nbin_MS = nbins

        # Single number divided equally between break masses
        if isinstance(nbin_MS, int):
            self._nbin_MS_each = _divide_bin_sizes(nbins, N_MS_breaks)

        # List of bins between each break mass
        else:
            self._nbin_MS_each = nbin_MS
            nbin_MS = np.sum(nbin_MS)

        # ------------------------------------------------------------------
        # Setup stellar mass bins (based entirely on edges)
        # ------------------------------------------------------------------

        # Equal-log-space between bins, started again at each break
        if binning_method in ('default', 'split_log', 'log_split'):

            # one-liner required for different Neach and no repeating breaks
            bin_sides = np.r_[tuple(
                np.geomspace(m_break[i], m_break[i + 1],
                             self._nbin_MS_each[i] + 1)[(i > 0):]
                for i in range(len(self._nbin_MS_each))
            )]

        elif binning_method in ('linear_split', 'split_linear'):

            bin_sides = np.r_[tuple(
                np.linspace(m_break[i], m_break[i + 1],
                            self._nbin_MS_each[i] + 1)[(i > 0):]
                for i in range(len(self._nbin_MS_each))
            )]

        # TODO unsure how to implement uniform spaces while still hitting breaks
        # elif binning_method in ('log', 'logged'):
        # elif binning_method in ('linear'):

        else:
            mssg = f"Unrecognized binning method '{binning_method}'"
            raise ValueError(binning_method)

        # Define a bin based on it's upper and lower bounds
        bins_MS = mbin(bin_sides[:-1], bin_sides[1:])

        # ------------------------------------------------------------------
        # Setup Remnant mass bins
        # ------------------------------------------------------------------

        # Nbins are specified in dict for each type, I guess log space them
        if isinstance(nbins, collections.abc.Mapping):

            if binning_method in ('default', 'split_log', 'log_split'):
                binfunc = np.geomspace
            else:
                binfunc = np.linspace

            # White Dwarfs

            nbin_WD = nbins['WD']

            WD_bl, WD_bu = ifmr.WD_mf
            WD_bl = m_break[0] if WD_bl < m_break[0] else WD_bl

            bin_sides = binfunc(WD_bl, WD_bu, nbin_WD)
            bins_WD = mbin(bin_sides[:-1], bin_sides[1:])

            # Black Holes

            nbin_BH = nbins['BH']

            BH_bl, BH_bu = ifmr.BH_mf
            BH_bu = m_break[-1] if BH_bu > m_break[-1] else BH_bu

            bin_sides = binfunc(BH_bl, BH_bu, nbin_BH)
            bins_BH = mbin(bin_sides[:-1], bin_sides[1:])

            # Neutron Stars

            nbin_NS = nbins.get('NS', 1)

            if nbin_NS != 1:
                mssg = "All NS have same mass, cannot have more than 1 bin"
                raise ValueError(mssg)

            # arbitrarily small width around 1.4
            bins_NS = mbin(*(np.array(ifmr.NS_mf) + [-0.01, 0.01]))

        # Divide out the bins as if they were cut out from the MS bins
        else:

            # White Dwarfs

            WD_mask = (bins_MS.lower <= ifmr.WD_mf.upper)

            nbin_WD = WD_mask.sum()

            # TODO fails if m_break[0] < ifmr.WD_mf.upper
            bins_WD = mbin(bins_MS.lower[WD_mask].copy(),
                           bins_MS.upper[WD_mask].copy())
            bins_WD.upper[-1] = ifmr.WD_mf.upper

            # Black Holes

            BH_mask = (bins_MS.upper > ifmr.BH_mf.lower)

            nbin_BH = BH_mask.sum()

            bins_BH = mbin(bins_MS.lower[BH_mask].copy(),
                           bins_MS.upper[BH_mask].copy())
            bins_BH.lower[0] = ifmr.BH_mf.lower

            # Neutron Stars

            NS_mask = (bins_MS.lower < 1.4) & (1.4 < bins_MS.upper)

            nbin_NS = NS_mask.sum()  # Always = 1

            # Doesn't really matter, but keep the full bin width here
            bins_NS = mbin(bins_MS.lower[NS_mask].copy(),
                           bins_MS.upper[NS_mask].copy())

        # ------------------------------------------------------------------
        # Setup the "blueprint" for the packed values
        # The blueprint is an array of integers reflecting the slices in the
        # packed `y` representing each component, to be used in `np.split(y)`.
        # Splitting includes the last given index to the end, so the final
        # nbin_BH should not be included in `_blueprint`, but is part of the
        # total size of y
        # ------------------------------------------------------------------

        self._blueprint = np.cumsum([0, nbin_MS, nbin_MS,  # Ms, alphas
                                     nbin_WD, nbin_NS, nbin_BH,  # N rem
                                     nbin_WD, nbin_NS, nbin_BH])  # M rem

        self._ysize = self._blueprint[-1]

        # ------------------------------------------------------------------
        # Save collections of useful attributes
        # ------------------------------------------------------------------

        self.nbin = star_classes(MS=nbin_MS, WD=nbin_WD, NS=nbin_NS, BH=nbin_BH)
        self.nbin_tot = np.r_[self.nbin].sum()

        self.bins = star_classes(MS=bins_MS, WD=bins_WD, NS=bins_NS, BH=bins_BH)

    def initial_values(self, *, packed=True):
        '''Return an array corresponding to `y` populated based on the IMF

        Based on the input IMF slopes and break masses, create an empty `y`
        array and populate it with "initial" values.
        These values will populate the number of stars and the alpha slopes,
        and all remnants will be left as 0.

        Parameters
        ----------
        packed : bool, optional
            If True (default) will return a single array, otherwise will return
            two named tuples of `star_classes` with the masses and numbers in
            each stellar class. Note that this is *not* the same as an unpacked
            array.

        Return
        ------
        ndarray or 2-tuple of named `star_classes` tuples
        '''

        # TODO this is where we restrict to 3comp IMF,
        #   but should be generalized to any number of comps really

        # ------------------------------------------------------------------
        # Compute normalization factors A_j
        # ------------------------------------------------------------------

        a, mb = self.a, self.m_break

        A3 = (
            Pk(a[2], 1, mb[2], mb[3])
            + (mb[2] ** (a[2] - a[1]) * Pk(a[1], 1, mb[1], mb[2]))
            + (mb[1] ** (a[1] - a[0]) * (mb[2] ** (a[2] - a[1]))
               * Pk(a[0], 1, mb[0], mb[1]))
        )**(-1)

        A2 = A3 * mb[2]**(a[2] - a[1])
        A1 = A2 * mb[1]**(a[1] - a[0])

        A = self.N0 * np.repeat([A1, A2, A3], self._nbin_MS_each)

        # ------------------------------------------------------------------
        # Set the initial Nj and mj for all bins (stars and remnants)
        # ------------------------------------------------------------------

        # Expand array of IMF slopes to all mass bins
        alpha = np.repeat(a, self._nbin_MS_each)

        # Set initial star bins based on IMF
        Ns = A * Pk(alpha, 1, *self.bins.MS)

        # Set all initial remnant bins to zero
        Nwd, Mwd = np.zeros(self.nbin.WD), np.zeros(self.nbin.WD)
        Nns, Mns = np.zeros(self.nbin.NS), np.zeros(self.nbin.NS)
        Nbh, Mbh = np.zeros(self.nbin.BH), np.zeros(self.nbin.BH)

        if packed:
            return self.pack_values(Ns, alpha, Nwd, Nns, Nbh, Mwd, Mns, Mbh)

        else:
            # this way so wouldn't have to recompute Ms from typical unpacked
            Ms = A * Pk(alpha, 2, *self.bins.MS)

            N = star_classes(MS=Ns, WD=Nwd, NS=Nns, BH=Nbh)
            M = star_classes(MS=Ms, WD=Mwd, NS=Mns, BH=Mbh)

            return N, M

    def blanks(self, value=0., extra_dims=None, *, packed=True, **kwargs):
        '''Returns an array corresponding to `y` with a single (blank) value

        Returns a new array (or set of arrays) corresponding to the `y`
        array, with all values set to a single given value, meant to represent
        initial empty arrays to be filled with values in the course of a
        derivative function.

        Any value is supported, but 0 or an empty array will be the fastest.
        If desired, the `y` array can unpacked into a list of arrays, such
        as is done by the `unpack_values` method.

        Multidimensional arrays can be returned by supplying extra dimensions.
        The final dimension will always be the required size (nbin_tot * 2).
        Be aware that, while `unpack_values` will support such a
        multidimensional `y` array as input, `pack_values` will not accept any
        multidimensional arrays.

        Parameters
        ----------
        value : float or 'empty', optional
            The value to fill the new blank array with. If 'empty', will use the
            `np.empty` initilization. The default (0) and 'empty' will be the
            most efficient, however any value is supported.

        extra_dims : None or list of int, optional
            An optional list of integer dimensions. If given, the shape of the
            final array will be (*extra_dims, nbin_tot * 2).

        packed : bool, optional
            If True (default) will return a single blank array, otherwise this
            will be passed to `unpack_values` first. All other kwargs will
            also be passed to this call.

        Returns
        -------
        ndarray or list of ndarray
            The final, blank, array with the given value
        '''

        shape = (*([] if extra_dims is None else extra_dims), self._ysize)

        if value == 0.:
            full = np.zeros(shape=shape)  # Much faster
        elif value == 'empty':
            full = np.empty(shape=shape)
        else:
            full = np.full(shape=shape, fill_value=value)

        return full if packed else self.unpack_values(full, **kwargs)

    def pack_values(self, Ns, alpha, Nwd, Nns, Nbh, Mwd, Mns, Mbh):
        '''Pack various arrays into a single `y` array

        Given 8 separate arrays, combine them into a single unified "packed"
        array, such as those required for input/output to numerical DE solvers.

        Parameters
        ----------
        Ns : ndarray
            Array containing values corresponding to the number of stars,
            placed first in the packed array. Must have size `nbin.MS`.

        alpha : ndarray
            Array containing values corresponding to stellar alphas
            placed second in the packed array. Must have size `nbin.MS`.

        Nwd : ndarray
            Array containing values corresponding to number of WD
            placed third in the packed array. Must have size `nbin.WD`.

        Nns : ndarray
            Array containing values corresponding to number of NS
            placed fourth in the packed array. Must have size `nbin.NS`.

        Nbh : ndarray
            Array containing values corresponding to number of BH
            placed fifth in the packed array. Must have size `nbin.BH`.

        Mwd : ndarray
            Array containing values corresponding to mass of WD
            placed sixth in the packed array. Must have size `nbin.WD`.

        Mns : ndarray
            Array containing values corresponding to mass of NS
            placed seventh in the packed array. Must have size `nbin.NS`.

        Mbh : ndarray
            Array containing values corresponding to mass of BH
            placed last in the packed array. Must have size `nbin.BH`.

        Returns
        -------
        y : ndarray
            Single "packed" array of values with total size nbin_tot * 2.

        See also
        --------
        unpack_values : The inverse of this method
        '''

        # Manually pack the arrays instead of simply concatenating
        # (i.e. `np.r_`), just to speed up the packing slightly

        inp_arrays = (Ns, alpha, Nwd, Nns, Nbh, Mwd, Mns, Mbh)
        out = self.blanks('empty')
        bp = self._blueprint
        for inp, i, j in zip(inp_arrays, bp[:-1], bp[1:]):
            out[i:j] = inp

        return out

    def unpack_values(self, y, *, grouped_rem=False):
        '''Unpack a given `y` array into the various expected arrays

        Taking in a single unified "packed" array, such as those passed into
        derivative functions from numerical DE solvers, splits the array into
        the 8 expected arrays (Ns, alpha, Nwd, Nns, Nbh, Mwd, Mns, Mbh) based
        on the number of bins in each object type, returning a list of all
        arrays.

        Should work for multidimensional `y` arrays, as long as the *final*
        dimension is the size of the total arrays (nbin_tot * 2).

        Parameters
        ----------
        y : ndarray
            Single "packed" array of values to unpack. Final (-1) dimension
            must have correct size (nbin_tot * 2).

        grouped_rem : bool, optional
            Whether to pack the arrays for the remnant numbers and masses
            (i.e. the last six returned arrays) into a single named tuple
            (`rem_classes`) instead of returning a full separated list of all
            arrays (the default).

        Returns
        -------
        list of ndarray
            A list of 8 arrays representing each sliced out section of the input
            array, with sizes based on the bin sizes of the relevant classes.
            If `grouped_rem` is True, the list will instead contain 4 elements,
            two arrays, exactly as usual, and 2 `rem_classes` named tuples each
            containing half of the 6 other arrays.

        See also
        --------
        pack_values : The inverse of this method
        '''

        # Manually unpack each array, rather than using `np.split`, based on
        # the blueprint, just to speed up the unpacking slightly.

        bp = self._blueprint
        out = [y[..., i:j] for i, j in zip(bp[:-1], bp[1:])]

        # If desired, group the remnants (M and N) into `rem_classes`
        if grouped_rem:
            Ns, alpha = out[0:2]
            Nrem = rem_classes(*out[2:5])
            Mrem = rem_classes(*out[5:8])
            return [Ns, alpha, Nrem, Mrem]

        else:
            return out

    def determine_index(self, mass, massbins, *, allow_overflow=False):
        '''Determine the index of the bin in `massbins` containing `mass`

        Given a set of massbins, determine the index of the bin which the given
        mass falls into.

        Note that mass bins are considered "left-inclusive" here,
        i.e. [lower, upper), therefore masses falling on the bin edges will
        return the lower index.

        Currently this function only supports one input mass at a time.

        Parameters
        ----------
        mass : float
            The mass from which to determine the location/index within the mass
            bins.

        massbins : mbin or {'MS', 'WD', 'NS', 'BH'}
            The massbins to determine the mass index within. Either a mass bin
            named tuple with upper and lower bound fields, or a two character
            string indicating the stellar type, which will pull from the
            massbins stored in this instance.
        '''
        # TODO support multiple mass at once, by correctly arranging mass shape

        # If a string labelling the class, get from stored bins
        # Otherwise assume it's already a massbins class
        if massbins in star_classes._fields:
            massbins = getattr(self.bins, massbins)

        # Since mass bins always increasing, can look at only the lower bound
        try:
            ind = np.flatnonzero(massbins.lower <= mass)[-1]
        except IndexError:
            mssg = f"mass {mass} is below lowest bound in {massbins}"
            raise ValueError(mssg)

        # If this is the last bin, maybe check if its actually overflowing
        if ind >= massbins.upper.size - 1:
            if (massbins.upper[-1] <= mass) and allow_overflow is False:
                mssg = f"mass {mass} is above highest bound in {massbins}"
                raise ValueError(mssg)

        return ind

    def turned_off_bins(self, mto):
        '''Return a copy of the MS bins with upper bounds adjusted up to `mto`

        Returns a copy of the stellar (MS) mass bins where the upper bounds of
        one bin, that holding the given turn-off mass (`mto`), are adjusted to
        fall on the `mto`. This may be required at times to, basically, spoof
        the fact that all mass within a given bin but above `mto` may have been
        lost (e.g. turned into remnants), and thus shouldn't be considered in
        some cases for computing quantities such as the mean bin mass.

        Parameters
        ----------
        mto : float
            The "turn-off" mass used to locate and truncate the turn-off bin.

        Returns
        -------
        mbin
            Near identical copy of the MS mass bin tuple, with the turn-off bin
            adjusted.
        '''

        low, up = self.bins.MS.lower.copy(), self.bins.MS.upper.copy()

        # Find the bin containing mto and adjust its upper bound only
        try:
            isev = self.determine_index(mto, 'MS')
            up[isev] = mto

        # If the mto is above or below the min/max bounds, just ignore it
        except ValueError:
            pass

        return mbin(low, up)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pathlib
import logging
import collections

import numpy as np
from scipy.interpolate import UnivariateSpline


__all__ = ["IFMR", "get_data"]


bounds = collections.namedtuple('bounds', ('lower', 'upper'))

_ROOT = pathlib.Path(__file__).parent


def get_data(path):
    '''Get path of data from path relative to install dir.'''
    return _ROOT / "data" / path


class IFMR:
    '''
    Provides a class for the initial-final mass of all stellar remnants.
    These are based on MIST and SSE models at different metallicities.


    mBH_min : float
        Alias to BH_mf[0], for backwards compatibility

    mWD_max : float
        Alias to WD_mf[1], for backwards compatibility
    '''

    def __init__(self, FeH):

        # ------------------------------------------------------------------
        # Check metallicity bounds
        # ------------------------------------------------------------------

        self.FeH = FeH

        self._check_feh_bounds()

        # ------------------------------------------------------------------
        # Black Holes
        # ------------------------------------------------------------------

        # TODO if 5e-3 < FeH < 0.0, this will put wrong sign on filename
        bhifmr = np.loadtxt(get_data(f"sse/MP_FEH{self.FeH_BH:+.2f}.dat"))

        # Grab only stellar type 14
        BH_mi, BH_mf = bhifmr[:, :2][bhifmr[:, 2] == 14].T

        # linear spline to avoid boundary effects near m_A, m_B, etc
        self._BH_spline = UnivariateSpline(BH_mi, BH_mf, s=0, k=1, ext=0)

        self.BH_mi = bounds(BH_mi[0], BH_mi[-1])

        # self.mBH_min, self.mBH_max = np.min(BH_mf), np.max(BH_mf)
        self.BH_mf = bounds(np.min(BH_mf), np.inf)
        self.mBH_min = self.BH_mf.lower

        # ------------------------------------------------------------------
        # White Dwarfs
        # ------------------------------------------------------------------

        wdgrid = np.loadtxt(get_data("sevtables/wdifmr.dat"))

        # Get the closest model
        j = np.argmin(np.abs(self.FeH_WD - wdgrid[:, 0]))
        WD_m_max, WD_coeffs = wdgrid[j, 1], wdgrid[j, 2:]

        self._WD_spline = np.polynomial.Polynomial(WD_coeffs[::-1])

        # TODO polynomial starts misbehaving far above 0, but don't know where
        self.WD_mi = bounds(0.0, WD_m_max)

        self.WD_mf = bounds(0.0, self._get_WD_bounds(WD_m_max))
        self.mWD_max = self.WD_mf.upper

        # ------------------------------------------------------------------
        # Neutron Stars
        # ------------------------------------------------------------------

        self.NS_mi = bounds(self.WD_mi[1], self.BH_mi[0])
        self.NS_mf = bounds(1.4, 1.4)

    def _check_feh_bounds(self):
        '''Ensure FeH is within model grid and adjust otherwise'''

        mssg = ("{0:.2f} is out of bounds for the {1} metallicity grid, "
                "falling back to {2} of {3:.2f}")

        # ------------------------------------------------------------------
        # White Dwarfs
        # ------------------------------------------------------------------

        wdgrid = np.loadtxt(get_data("sevtables/wdifmr.dat"))[:, 0]

        if self.FeH < (fback := np.min(wdgrid)):
            logging.debug(mssg.format(self.FeH, 'WD', 'minimum', fback))
            self.FeH_WD = fback

        elif self.FeH > (fback := np.max(wdgrid)):
            logging.debug(mssg.format(self.FeH, 'WD', 'maximum', fback))
            self.FeH_WD = fback

        else:
            self.FeH_WD = self.FeH

        # ------------------------------------------------------------------
        # Black Holes
        # ------------------------------------------------------------------

        bhgrid = np.array([float(fn.stem.split('FEH')[-1])
                           for fn in get_data('sse').glob('MP*dat')])

        if self.FeH < (fback := np.min(bhgrid)):
            logging.debug(mssg.format(self.FeH, 'BH', 'minimum', fback))
            self.FeH_BH = fback

        elif self.FeH > (fback := np.max(bhgrid)):
            logging.debug(mssg.format(self.FeH, 'BH', 'maximum', fback))
            self.FeH_BH = fback

        else:
            self.FeH_BH = self.FeH

    def _get_WD_bounds(self, max_mi):
        '''compute the WD_mf bounds as best we can from the polynomial'''

        # Compute the max/mins taking derivative of polynomial
        WD_minmax = self._WD_spline.deriv().roots().real

        # Restrict x (initial mass) to between 1 and max_mi
        # TODO lower 1 is required to avoid polynomial effects, but is arbitrary
        #   really it won't matter, maximum will always be far above 1
        restr = (1. < WD_minmax) & (WD_minmax <= max_mi)

        # Determine the maximum WD final mass (including upper bound in case)
        return self._WD_spline(np.r_[WD_minmax[restr], max_mi]).max()

    def predict_type(self, m_in):
        '''Predict the remnant type (WD, NS, BH) given the initial mass(es)'''

        rem_type = np.where(
            m_in >= self.BH_mi[0], 'BH',
            np.where(
                (self.WD_mi[1] < m_in) & (m_in <= self.BH_mi[0]), 'NS',
                'WD'
            )
        )

        return rem_type.tolist()

    def predict(self, m_in):
        '''Predict the final mass given the initial mass(es) `m_in`'''

        final = np.where(
            m_in >= self.BH_mi[0], self._BH_spline(m_in),
            np.where(
                (self.WD_mi[1] < m_in) & (m_in <= self.BH_mi[0]), 1.4,
                self._WD_spline(np.array(m_in))
            )
        )

        # If outside boundaries of the IFMR, warn user
        if np.any((m_in <= self.WD_mi[0]) | (m_in > self.BH_mi[1])):
            mssg = ("input mass exceeds IFMR grid, resulting mass is "
                    "extrapolated and may be very incorrect")
            logging.warning(mssg)

        # if m_in is a single float, reconvert to match
        if not final.shape:
            final = float(final)

        return final



#!/usr/bin/env python

import pytest
import numpy as np

from ssptools import masses, ifmr


DEFAULT_KWARGS = dict(
    m_break=[0.1, 0.5, 1.0, 100], a=[-0.5, -1.3, -2.5],
    nbins=[1, 1, 2], N0=5e5, ifmr=ifmr.IFMR(-1)
)


class TestHelperMethods:

    # ----------------------------------------------------------------------
    # Testing division of bins helper function
    # ----------------------------------------------------------------------

    @pytest.mark.parametrize(
        "N, Nsec, expected",
        [
            (10, 2, [5, 5]),
            (10, 3, [4, 3, 3]),
            (2, 5, [1, 1, 0, 0, 0]),
            (0, 5, [0, 0, 0, 0, 0]),
            (10, 0, pytest.raises(ZeroDivisionError)),
            (0, 0, pytest.raises(ZeroDivisionError))
        ]
    )
    def test_divide_bin_sizes(self, N, Nsec, expected):

        if isinstance(expected, list):
            assert masses._divide_bin_sizes(N, Nsec) == expected

        else:
            with expected:
                masses._divide_bin_sizes(N, Nsec)

    # ----------------------------------------------------------------------
    # Testing computation of P_k helper integral solution
    # ----------------------------------------------------------------------

    @pytest.mark.parametrize('k', [1., 1.5, 2.])
    @pytest.mark.parametrize('a', [-2, -1., -0.5, 1.0])
    def test_Pk(self, a, k):
        from scipy.integrate import quad

        m1, m2 = 0.5, 1.0
        expected, err = quad(lambda m: m**(a + k - 1), m1, m2)

        Pk = masses.Pk(a=a, k=k, m1=m1, m2=m2)

        assert Pk == pytest.approx(expected, abs=err)


class TestArrayPacking:

    # Test the packing and unpacking of y

    mb = masses.MassBins(**DEFAULT_KWARGS)

    # ----------------------------------------------------------------------
    # Test the "packing" and "unpacking" functionality for ODE style y-arrays
    # ----------------------------------------------------------------------

    def test_packing(self):
        arrays = np.split(range(self.mb._ysize), self.mb._blueprint[1:-1])

        res = self.mb.pack_values(*arrays)

        expected = np.concatenate(arrays)

        np.testing.assert_equal(res, expected)

    def test_unpacking(self):

        y = np.arange(self.mb._ysize)

        res = self.mb.unpack_values(y)

        expected = np.split(range(self.mb._ysize), self.mb._blueprint[1:-1])

        assert len(res) == len(expected)

        for r, e in zip(res, expected):
            np.testing.assert_equal(r, e)


class TestArrayCreation:

    # Test the creation of both initial and blank arrays

    mb = masses.MassBins(**DEFAULT_KWARGS)

    # ----------------------------------------------------------------------
    # Testing creation of "blanks" of correct sizes
    # ----------------------------------------------------------------------
    # Parametrize with values each time because internal logic differs w/ value
    # Use numpy.testing because we want exact matches, not pytest.approx

    @pytest.mark.parametrize('value', [1., 0.])  # TODO how to test empty?
    def test_blanks_value(self, value):

        kw = {'extra_dims': None, 'packed': True}

        res = self.mb.blanks(value=value, **kw)

        expected = np.array([value] * self.mb._ysize)

        np.testing.assert_equal(res, expected)

    @pytest.mark.parametrize('value', [1., 0.])
    @pytest.mark.parametrize('dims', [[], [1], [1, 2, 3]])
    def test_blanks_shape(self, value, dims):

        kw = {'packed': True}

        res = self.mb.blanks(value=value, extra_dims=dims, **kw)

        expected = np.full(fill_value=value, shape=(*dims, self.mb._ysize))

        np.testing.assert_equal(res, expected)

    @pytest.mark.parametrize('value', [1., 0.])
    @pytest.mark.parametrize('packed', [True, False])
    def test_blanks_packing(self, value, packed):

        kw = {'extra_dims': None}

        res = self.mb.blanks(value=value, packed=packed, **kw)

        if packed is True:
            expected = np.array([value] * self.mb._ysize)

            np.testing.assert_equal(res, expected)

        else:
            nbin = self.mb.nbin
            shape = (nbin.MS, nbin.MS, nbin.WD, nbin.NS,
                     nbin.BH, nbin.WD, nbin.NS, nbin.BH)
            expected = [np.array([value] * i) for i in shape]

            assert len(res) == len(expected)

            for r, e in zip(res, expected):
                np.testing.assert_equal(r, e)


class TestMassIndices:
    from contextlib import nullcontext as does_not_raise

    # Test the finding of indices and of changing it based on mto

    mb = masses.MassBins(**DEFAULT_KWARGS)

    # ----------------------------------------------------------------------
    # Test determining the index of a given mass
    # ----------------------------------------------------------------------

    @pytest.mark.parametrize(
        "mass, massbin, expected",
        [
            (0.99, 'MS', 1),
            (1., 'MS', 2),
            (1., 'WD', 2),
            (1., 'NS', 0),
            (DEFAULT_KWARGS['ifmr'].BH_mf.lower, 'BH', 0),
            (99., 'BH', 1),
            (2., masses.mbin(*np.array([[1, 2, 3], [2, 3, 4]])), 1)
        ]
    )
    def test_determine_index(self, mass, massbin, expected):

        assert self.mb.determine_index(mass, massbin) == expected

    @pytest.mark.parametrize(
        "value, overflow, expected",
        [
            (0.01, True, pytest.raises(ValueError)),
            (0.1, True, does_not_raise()),
            (100, True, does_not_raise()),
            (101, True, does_not_raise()),
            #
            (0.01, False, pytest.raises(ValueError)),
            (0.1, False, does_not_raise()),
            (100, False, pytest.raises(ValueError)),
            (101, False, pytest.raises(ValueError)),
        ]
    )
    def test_determine_overflow(self, value, overflow, expected):

        with expected:
            self.mb.determine_index(value, 'MS', allow_overflow=overflow)

    # ----------------------------------------------------------------------
    # Test rearranging the mass bins based on a given mto
    # ----------------------------------------------------------------------

    @pytest.mark.parametrize(
        "mto, expected_upper",
        [
            (-1, [0.5, 1, 10, 100]),
            (0, [0.5, 1, 10, 100]),
            (0.3, [0.3, 1, 10, 100]),
            (0.5, [0.5, 0.5, 10, 100]),
            (101, [0.5, 1, 10, 100])
        ]
    )
    def test_to_bins(self, mto, expected_upper):

        res = self.mb.turned_off_bins(mto).upper

        np.testing.assert_equal(res, expected_upper)


# class TestMassBinInit:

    # Test the creation of mass bins with various methods



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from scipy.integrate import ode
from scipy.interpolate import interp1d, UnivariateSpline

from .ifmr import IFMR, get_data
from .masses import MassBins, Pk, mbin


# TODO optionally support units for some things


__all__ = ['EvolvedMF', 'EvolvedMFWithBH', 'InitialBHPopulation']


class EvolvedMF:
    r'''Evolve an IMF to a present-day mass function at a given age.

    Evolves an arbitrary N-component power law initial mass function (IMF)
    to a binned present-day mass function (PDMF) at a given set of ages, and
    computes the numbers and masses of stars and remnants in each mass bin.

    # TODO add more in-depth explanation or references to actual algorithm here

    Warning: Currently only supports N=3 IMF components.

    Parameters
    ----------
    m_breaks : list of float
        IMF break-masses (including outer bounds; size N+1).

    a_slopes : list of float
        IMF slopes α (size N).

    nbins : list of int
        Number of mass bins in each regime of the IMF, as defined by m_breaks
        (size N).

    FeH : float
        Metallicity, in solar fraction [Fe/H].

    tout : list of int
        Times, in years, at which to output PDMF. Defines the shape of many
        outputted attributes.

    Ndot : float
        Represents rate of change of the number of stars N over time, in stars
        per Myr. Regulates low-mass object depletion (ejection) due to dynamical
        evolution.

    N0 : int, optional
        Total initial number of stars, over all bins. Defaults to 5e5 stars.

    tcc : float, optional
        Core collapse time, in years. Defaults to 0, effectively being ignored.

    NS_ret : float, optional
        Neutron star retention fraction (0 to 1). Defaults to 0.1 (10%).

    BH_ret_int : float, optional
        Initial black hole retention fraction (0 to 1). Defaults to 1 (100%).

    BH_ret_dyn : float, optional
        Dynamical black hole retention fraction (0 to 1), including both
        dynamical ejections and natal kicks. Defaults to 1 (100%).

    natal_kicks : bool, optional
        Whether to account for natal kicks in the BH dynamical retention.
        Defaults to False.

    vesc : float, optional
        Initial cluster escape velocity, in km/s, for use in the computation of
        BH natal kick effects. Defaults to 90 km/s.

    binning_method : {'default', 'split_log', 'split_linear'}, optional
        The spacing method to use when constructing the mass bins.
        See `masses.MassBins` for more information.

    md : float, optional
        Depletion mass, below which stars are preferentially disrupted during
        the stellar escape derivatives.
        The default 1.2 is based on Lamers et al. (2013) and shouldn't be
        changed unless you know what you're doing.

    Attributes
    ----------

    nbin : int
        Total number of bins (sum of nbins parameter).

    nout : int
        Total number of output times (len of tout parameter).

    alphas : ndarray
        Array[nout, nbin] of  PDMF slopes. If Ndot = 0, this is defined entirely
        by the IMF.

    Ns : ndarray
        Array[nout, nbin] of the total number of (main sequence) stars in each
        mass bin.

    Ms : ndarray
        Array[nout, nbin] of the total mass of (main sequence) stars in each
        mass bin.

    ms : ndarray
        Array[nout, nbin] of the individual mass of (main sequence) stars in
        each mass bin, as defined by Ms / Ns.

    mes : ndarray
        Array[nout, nbin + 1] representing the mass bin edges, defined by the
        m_breaks and nbins parameters. Note that ms *does not* necessarily fall
        in the middle of these edges.

    Nr : ndarray
        Array[nout, nbin] of the total number of remnants in each mass bin.

    Mr : ndarray
        Array[nout, nbin] of the total mass of remnants in each mass bin.

    mr : ndarray
        Array[nout, nbin] of the individual mass of remnants in
        each mass bin, as defined by Mr / Nr.

    rem_types : ndarray
        Array[nout, nbin] of 2-character strings representing the remnant types
        found in each mass bin of the remnant (Mr, Nr, mr) arrays. "WD" = white
        dwarfs, "NS" = neutron stars, "BH" = black holes.

    M : ndarray
        Array containing the total mass in all non-empty mass bins, including
        both stars and remnants, based on the final age given by tout.

    N : ndarray
        Array containing the total number in all non-empty mass bins, including
        both stars and remnants, based on the final age given by tout.

    m : ndarray
        Array containing the individual mass in all non-empty mass bins,
        including both stars and remnants, based on the final age given by tout.

    bin_widths : ndarray
        Array containing the width of all non-empty mass bins,
        including both stars and remnants, based on the final age given by tout.
        Equal to `np.diff(mes)`.

    types : ndarray
        Array of 2-character strings representing the object type of the
        corresponding bins of the M, N and m properties. "MS" = main sequence
        stars, "WD" = white dwarfs, "NS" = neutron stars, "BH" = black holes.
    '''

    @property
    def M(self):
        # TODO why is it 10*Nmin
        cs = self.Ns[-1] > 10 * self.Nmin
        Ms = self.Ms[-1][cs]

        cr = np.c_[self.Nr][-1] > 10 * self.Nmin
        Mr = np.c_[self.Mr][-1][cr]

        return np.r_[Ms, Mr]

    @property
    def N(self):
        cs = self.Ns[-1] > 10 * self.Nmin
        Ns = self.Ns[-1][cs]

        cr = np.c_[self.Nr][-1] > 10 * self.Nmin
        Nr = np.c_[self.Nr][-1][cr]

        return np.r_[Ns, Nr]

    @property
    def m(self):
        return self.M / self.N

    @property
    def bin_widths(self):

        mto = self.compute_mto(self.tout[-1])
        mes = self.massbins.turned_off_bins(mto)

        cs = self.Ns[-1] > 10 * self.Nmin
        bws = (mes.upper - mes.lower)[cs]

        cr = np.c_[self.Nr][-1] > 10 * self.Nmin
        bwr = np.diff(np.c_[self.massbins.bins[1:]], axis=0)[-1][cr]

        return np.r_[bws, bwr]

    @property
    def types(self):
        cs = self.Ns[-1] > 10 * self.Nmin
        cr = np.c_[self.Nr][-1] > 10 * self.Nmin

        ts = ['MS'] * cs.sum()
        tr = self.rem_types[cr]

        return np.r_[ts, tr]

    @property
    def nms(self):
        return (self.Ns[-1] > 10 * self.Nmin).sum()

    @property
    def nmr(self):
        return (np.c_[self.Nr][-1] > 10 * self.Nmin).sum()

    def __init__(self, m_breaks, a_slopes, nbins, FeH, tout, Ndot,
                 N0=5e5, tcc=0.0, NS_ret=0.1, BH_ret_int=1.0, BH_ret_dyn=1.0,
                 natal_kicks=False, vesc=90, binning_method='default', md=1.2):

        # ------------------------------------------------------------------
        # Set various other parameters
        # ------------------------------------------------------------------

        # Supplied parameters
        self.tcc = tcc
        self.tout = np.atleast_1d(tout)
        self.Ndot = Ndot

        self.NS_ret = NS_ret
        self.BH_ret_int = BH_ret_int
        self.BH_ret_dyn = BH_ret_dyn
        self._frem = {'WD': 1., 'NS': NS_ret, 'BH': BH_ret_int}

        self.FeH = FeH

        if Ndot > 0:
            raise ValueError("'Ndot' must be less than 0")

        # Initial-Final mass relations
        self.IFMR = IFMR(FeH)

        # Minimum of stars to call a bin "empty"
        self.Nmin = 0.1

        self.md = md

        # ------------------------------------------------------------------
        # Initialise the mass bins given the power-law IMF slopes and bins
        # ------------------------------------------------------------------

        self.massbins = MassBins(m_breaks, a_slopes, nbins, N0, self.IFMR,
                                 binning_method=binning_method)

        # ------------------------------------------------------------------
        # Setup lifetime approximations and compute t_ms of all bin edges
        # ------------------------------------------------------------------

        # Load a_i coefficients derived from interpolated Dartmouth models
        mstogrid = np.loadtxt(get_data("sevtables/msto.dat"))
        nearest_FeH = np.argmin(np.abs(mstogrid[:, 0] - FeH))
        self._tms_constants = mstogrid[nearest_FeH, 1:]

        # Compute t_ms for all bin edges
        self.tms_l = self.compute_tms(self.massbins.bins.MS.lower)
        self.tms_u = self.compute_tms(self.massbins.bins.MS.upper)

        # ------------------------------------------------------------------
        # Generate times for integrator
        # ------------------------------------------------------------------

        t_end = np.max(tout)

        # Compute each time a new bin evolves out of MS, up till t_end
        self.t = np.sort(np.r_[self.tms_u[self.tms_u < t_end], self.tout])

        self.nt = self.t.size

        # ------------------------------------------------------------------
        # Setup BH natal kicks
        # ------------------------------------------------------------------

        self.vesc = vesc
        self.natal_kicks = natal_kicks

        if self.natal_kicks:

            # load in the ifmr data to interpolate fb from mr
            feh_path = get_data(f"sse/MP_FEH{self.IFMR.FeH_BH:+.2f}.dat")

            # load in the data
            self.fb_grid = np.loadtxt(feh_path, usecols=(1, 3), unpack=True)

        # ------------------------------------------------------------------
        # Finally, evolve the population
        # ------------------------------------------------------------------

        # Initialize iteration counter
        self.nstep = 0

        self._evolve()

    def compute_tms(self, mi):
        '''Compute main-sequence lifetime for a given mass `mi`'''
        a = self._tms_constants
        return a[0] * np.exp(a[1] * mi ** a[2])

    def compute_mto(self, t):
        '''Compute the turn-off mass for a given time `t` (inverse of tms)'''
        a0, a1, a2 = self._tms_constants

        out = np.empty_like(t, dtype=float)
        asympt = t > a0

        out[asympt] = (np.log(np.asanyarray(t)[asympt] / a0) / a1) ** (1 / a2)
        out[~asympt] = np.inf

        return out

    def _derivs(self, t, y):
        '''Main function for computing derivatives relevant to mass evolution

        Simply calls the two constituent mass evolution derivative methods;
        `_derivs_esc` and `_derivs_sev`. Designed to be solved using an ODE
        integrator, such as `scipy.integrate.ode`.

        Parameters
        ----------
        t : float
            Time step to compute derivatives at

        y : list of ndarray
            Equation system solution y. Size-4 array containing the arrays,
            for each mass bin, of the number of stars `Ns`, the mass function
            slopes `alpha`, the number of remnants `Nr` and the total bin mass
            of remnants `Mr`

        Returns
        -------
        list of ndarray
            Time derivatives of each of the four quantities described by `y`
        '''

        # Iterate step count
        self.nstep += 1

        # Compute stellar evolution derivatives
        derivs_sev = self._derivs_sev(t, y)

        # Only run the dynamical star losses `derivs_esc` if Ndot is not zero
        if self.Ndot < 0:
            derivs_esc = self._derivs_esc(t, y)
        else:
            derivs_esc = np.zeros_like(derivs_sev)

        # Combine mass loss derivatives
        return derivs_sev + derivs_esc

    def _derivs_sev(self, t, y):
        '''Derivatives relevant to mass changes due to stellar evolution'''

        # Setup derivative bins
        Ns, alpha, Nr, Mr = self.massbins.unpack_values(y, grouped_rem=True)
        dNs, dalpha, dNr, dMr = self.massbins.blanks(packed=False,
                                                     grouped_rem=True)

        # Apply only if this time is atleast later than the earliest tms
        if t > self.tms_u[-1]:

            # Find out which mass bin is the current turn-off bin
            isev = np.where(t > self.tms_u)[0][0]

            m1 = self.massbins.bins.MS.lower[isev]
            mto = self.compute_mto(t)
            Nj = Ns[isev]

            # Avoid "hitting" the bin edge
            if mto > m1 and Nj > self.Nmin:

                # Two parameters defining the bin
                alphaj = alpha[isev]

                # The normalization constant
                # TODO deal with the constant divide-by-zero warning here
                Aj = Nj / Pk(alphaj, 1, m1, mto)

                # Get the number of turn-off stars per unit of mass
                dNdm = Aj * mto ** alphaj

            else:
                dNdm = 0
                # TODO just break???

            # Compute the full dN/dt = dN/dm * dm/dt
            a = self._tms_constants
            dmdt = abs((1.0 / (a[1] * a[2] * t))
                       * (np.log(t / a[0]) / a[1]) ** (1 / a[2] - 1))

            dNdt = -dNdm * dmdt

            # Fill in star derivatives (alphaj remains constant for _derivs_sev)
            dNs[isev] = dNdt

            # Find remnant mass and which bin they go into
            m_rem, cls_rem = self.IFMR.predict(mto), self.IFMR.predict_type(mto)

            # Skip 0-mass remnants
            if m_rem > 0:

                # Find bin based on lower bin edge (must be careful later)
                irem = self.massbins.determine_index(m_rem, cls_rem)

                # Compute Remnant retention fractions based on remnant type
                frem = self._frem[cls_rem]

                # Fill in remnant derivatives
                getattr(dNr, cls_rem)[irem] = -dNdt * frem
                getattr(dMr, cls_rem)[irem] = -m_rem * dNdt * frem

        return self.massbins.pack_values(dNs, dalpha, *dNr, *dMr)

    def _derivs_esc(self, t, y):
        '''Derivatives relevant to mass loss due to escaping low-mass stars'''

        # nb = self.nbin
        md, Ndot = self.md, self.Ndot

        # Pull out individual arrays from y
        Ns, alpha, Nr, Mr = self.massbins.unpack_values(y, grouped_rem=True)
        dNs, dalpha, dNr, dMr = self.massbins.blanks(packed=False,
                                                     grouped_rem=True)

        # If core collapsed, use different, simpler, algorithm
        if t < self.tcc:

            N_sum = np.sum(Ns) + np.sum(np.r_[Nr])
            dNs += Ndot * Ns / N_sum

            for c in range(len(Nr)):  # Each remnant class
                sel = Nr[c] > 0
                mr = Mr[c][sel] / Nr[c][sel]
                dNr[c][sel] += Ndot * Nr[c][sel] / N_sum
                dMr[c][sel] += mr * Ndot * Nr[c][sel] / N_sum

            return self.massbins.pack_values(dNs, dalpha, *dNr, *dMr)

        # Stellar Integrals

        mto = self.compute_mto(t)

        bins_MS = self.massbins.turned_off_bins(mto)

        P1 = Pk(alpha, 1, *bins_MS)
        P15 = Pk(alpha, 1.5, *bins_MS)
        P2 = Pk(alpha, 2, *bins_MS)

        ms = P2 / P1
        depl_mask = ms < md

        Is = (Ns * (1 - md ** (-0.5) * (P15 / P1)))[depl_mask]

        # Remnant integrals

        *_, Ir, Jr = self.massbins.blanks(packed=False, grouped_rem=True)
        for c in range(len(Nr)):  # Each remnant class

            rem_mask = Nr[c] > 0

            mr = Mr[c][rem_mask] / Nr[c][rem_mask]

            Ir[c][rem_mask] = np.where(
                mr < md, Nr[c][rem_mask] * (1 - np.sqrt(mr / md)), 0
            )
            Jr[c][rem_mask] = np.where(
                mr < md, Mr[c][rem_mask] * (1 - np.sqrt(mr / md)), 0
            )

        # Normalization

        B = Ndot / (np.sum(Is) + np.sum(np.r_[Ir]))

        # Derivatives

        dNs[depl_mask] += B * Is

        dalpha[depl_mask] += (
            B * ((bins_MS.lower[depl_mask] / md) ** 0.5
                 - (bins_MS.upper[depl_mask] / md) ** 0.5)
            / np.log(bins_MS.upper[depl_mask] / bins_MS.lower[depl_mask])
        )

        for c in range(len(Nr)):  # Each remnant class

            rem_mask = Nr[c] > 0

            dNr[c][rem_mask] += B * Ir[c][rem_mask]
            dMr[c][rem_mask] += B * Jr[c][rem_mask]

        return self.massbins.pack_values(dNs, dalpha, *dNr, *dMr)

    def _get_retention_frac(self, fb, vesc):
        '''Compute BH natal-kick retention fraction'''

        def _maxwellian(x, a):
            norm = np.sqrt(2 / np.pi)
            exponent = (x ** 2) * np.exp((-1 * (x ** 2)) / (2 * (a ** 2)))
            return norm * exponent / a ** 3

        if fb == 1.0:
            return 1.0

        # Integrate over the Maxwellian up to the escape velocity
        v_space = np.linspace(0, 1000, 1000)

        # TODO might be a quicker way to numerically integrate than a spline
        retention = UnivariateSpline(
            x=v_space,
            y=_maxwellian(v_space, 265 * (1 - fb)),
            s=0,
            k=3,
        ).integral(0, vesc)

        return retention

    def _natal_kick_BH(self, Mr_BH, Nr_BH):
        '''Computes the effects of BH natal-kicks on the mass and number of BHs

        Determines the effects of BH natal-kicks based on the fallback fraction
        and escape velocity of this population.

        Both input arrays are modified *in place*, as well as returned.

        Parameters
        ----------
        Mr_BH : ndarray
            Array[nbin] of the total initial masses of black holes in each
            BH mass bin

        Nr_BH : ndarray
            Array[nbin] of the total initial numbers of black holes in each
            BH mass bin

        Returns
        -------
        Mr_BH : ndarray
            Array[nbin] of the total final masses of black holes in each
            BH mass bin, after natal kicks

        Nr_BH : ndarray
            Array[nbin] of the total final numbers of black holes in each
            BH mass bin, after natal kicks

        natal_ejecta : float
            The total mass of BHs ejected through natal kicks
        '''

        natal_ejecta = 0.0

        # Interpolate the mr-fb grid
        fb_interp = interp1d(
            self.fb_grid[0],
            self.fb_grid[1],
            kind="linear",
            bounds_error=False,
            fill_value=(0.0, 1.0),
        )

        for j in range(Mr_BH.size):

            # Skip the bin if its empty
            if Nr_BH[j] < self.Nmin:
                continue

            # Interpolate fallback fraction at this mr
            fb = fb_interp(Mr_BH[j] / Nr_BH[j])

            if fb == 1.0:
                continue

            else:

                # Compute retention fraction
                retention = self._get_retention_frac(fb, self.vesc)

                # keep track of how much we eject
                natal_ejecta += Mr_BH[j] * (1 - retention)

                # eject the mass
                Mr_BH[j] *= retention
                Nr_BH[j] *= retention

        return Mr_BH, Nr_BH, natal_ejecta

    def _dyn_eject_BH(self, Mr_BH, Nr_BH, *, M_eject=None):
        '''Determine and remove BHs, to represent dynamical ejections

        Determines and removes an amount of BHs from the BH mass bins, designed
        to reflect the effects of dynamical ejections of centrally concentrated
        BHs from a cluster.
        Proceeds from the heaviest BH bins to the lightest, removing all mass
        in each bin until the desired amount of ejected mass is reached.

        Both input arrays are modified *in place*, as well as returned.

        Parameters
        ----------
        Mr_BH : ndarray
            Array[nbin] of the total initial masses of black holes in each
            BH mass bin

        Nr_BH : ndarray
            Array[nbin] of the total initial numbers of black holes in each
            BH mass bin

        M_eject : float, optional
            Total amount of BH mass to remove through dynamical ejections.
            If None (default), will be computed based on the `BH_ret_dyn`
            parameter

        Returns
        -------
        Mr_BH : ndarray
            Array[nbin] of the total final masses of black holes in each
            BH mass bin, after dynamical ejections

        Nr_BH : ndarray
            Array[nbin] of the total final numbers of black holes in each
            BH mass bin, after dynamical ejections
        '''

        # calculate total mass we want to eject
        if M_eject is None:
            M_eject = Mr_BH.sum() * (1.0 - self.BH_ret_dyn)

        # Remove BH starting from Heavy to Light
        j = Mr_BH.size

        while M_eject != 0:
            j -= 1

            if j < 0:
                mssg = 'Invalid `M_eject`, must be less than total Mr_BH'
                raise ValueError(mssg)

            # Skip empty bins
            if Nr_BH[j] < self.Nmin:
                continue

            # Removed entirety of this bin
            if Mr_BH[j] < M_eject:
                M_eject -= Mr_BH[j]
                Mr_BH[j] = 0
                Nr_BH[j] = 0
                continue

            # Remove required fraction of the last affected bin
            else:

                mr_BH_j = Mr_BH[j] / Nr_BH[j]

                Mr_BH[j] -= M_eject
                Nr_BH[j] -= M_eject / (mr_BH_j)

                break

        return Mr_BH, Nr_BH

    def _evolve(self):
        '''Main population evolution function, to be called on init'''

        # ------------------------------------------------------------------
        # Initialize output arrays
        # ------------------------------------------------------------------

        self.nout = len(self.tout)

        # Stars

        self.Ns, self.alpha, self.Nr, self.Mr = self.massbins.blanks(
            'empty', extra_dims=[self.nout], packed=False, grouped_rem=True
        )

        self.Ms, self.ms = np.empty_like(self.Ns), np.empty_like(self.Ns)
        self.mr = self.Nr._make(np.empty_like(x) for x in self.Nr)

        self.rem_types = np.repeat(self.massbins.nbin._fields[1:],
                                   self.massbins.nbin[1:])

        # ------------------------------------------------------------------
        # Initialise ODE solver
        # ------------------------------------------------------------------

        t0 = 0.0
        y0 = self.massbins.initial_values()

        sol = ode(self._derivs)
        sol.set_integrator("dopri5", max_step=1e12, atol=1e-5, rtol=1e-5)
        sol.set_initial_value(y0, t=t0)

        # ------------------------------------------------------------------
        # Evolve
        # ------------------------------------------------------------------

        for ti in self.t:

            # --------------------------------------------------------------
            # Integrate ODE solver
            # --------------------------------------------------------------

            sol.integrate(ti)

            # --------------------------------------------------------------
            # if this time is in the desired output times extract solutions
            # --------------------------------------------------------------

            if ti in self.tout:

                iout = np.where(self.tout == ti)[0][0]

                # ----------------------------------------------------------
                # Extract the N, M and alphas for stars and remnants
                # ----------------------------------------------------------

                Ns, alpha, Nr, Mr = self.massbins.unpack_values(
                    sol.y, grouped_rem=True
                )

                bins_MS = self.massbins.turned_off_bins(self.compute_mto(ti))

                As = Ns / Pk(alpha, 1, *bins_MS)
                Ms = As * Pk(alpha, 2, *bins_MS)

                # ----------------------------------------------------------
                # Eject BHs, first through natal kicks, then dynamically
                # ----------------------------------------------------------
                # TODO really feels like this should be done during the
                #   evolution/derivs? At least for the natal kicks.
                # TODO due to rem_classes tuple, ejections done in place, which
                #   is not ideal

                # Check if any BH have been created
                if ti > self.compute_tms(self.IFMR.BH_mi.upper):

                    # calculate total mass we want to eject
                    M_eject = Mr.BH.sum() * (1.0 - self.BH_ret_dyn)

                    # First remove mass from all bins by natal kicks, if desired
                    if self.natal_kicks:
                        *_, kicked = self._natal_kick_BH(Mr.BH, Nr.BH)
                        M_eject -= kicked

                    # Remove dynamical BH ejections
                    self._dyn_eject_BH(Mr.BH, Nr.BH, M_eject=M_eject)

                # ----------------------------------------------------------
                # save values into output arrays
                # ----------------------------------------------------------

                self.alpha[iout, :] = alpha

                # Stars
                self.Ns[iout, :] = Ns
                self.Ms[iout, :] = Ms

                self.ms[iout, :] = Ms / Ns

                # Remnants
                for c in range(len(Nr)):  # Each remnant class

                    self.Nr[c][iout, :] = Nr[c]
                    self.Mr[c][iout, :] = Mr[c]

                    # Precise mr only matters when Nr > 0
                    mr = 0.5 * np.sum(self.massbins.bins[c + 1], axis=0)

                    rem_mask = Nr[c] > 0
                    mr[rem_mask] = Mr[c][rem_mask] / Nr[c][rem_mask]

                    self.mr[c][iout, :] = mr


class EvolvedMFWithBH(EvolvedMF):
    r'''Evolve an IMF to a present-day mass function at a given age and f_BH.

    Subclass of `EvolvedMF`, evolving an arbitrary power law initial mass
    function (IMF) to a binned present-day mass function (PDMF) at a given set
    of ages, with an alternative BH prescription, allowing for a specified
    final BH mass fraction `f_BH`, rather than a BH retention fraction
    dependant on the amount of BHs created from the IMF.

    The given `f_BH` must be a valid fraction, between 0-1, and must be less
    than the mass fraction of BHs formed initially from the (and optionally
    after natal kicks), as BHs can only be removed through dynamical ejections.

    Parameters
    ----------
    m_breaks : list of float
        IMF break-masses (including outer bounds; size N+1).

    a_slopes : list of float
        IMF slopes α (size N).

    nbins : list of int
        Number of mass bins in each regime of the IMF, as defined by m_breaks
        (size N).

    FeH : float
        Metallicity, in solar fraction [Fe/H].

    tout : list of int
        Times, in years, at which to output PDMF. Defines the shape of many
        outputted attributes.

    Ndot : float
        Represents rate of change of the number of stars N over time, in stars
        per Myr. Regulates low-mass object depletion (ejection) due to dynamical
        evolution.

    f_BH : float
        The desired final BH mass fraction (0 to 1).
        If more than one output time is specified, a list of floats with size
        `nout` is required.

    *args, **kwargs
        All remaining arguments are passed to `EvolvedMF`. Note that passing
        a `BH_ret_dyn` will have no effect on the BHs.

    Notes
    -----
    The final `f_BH` as determined by
    `self.M[self.types=='BH'].sum() / self.M.sum()` may be very slightly
    (ε < 1e-3) different from the requested `f_BH`, simply due to the removal
    of the nearly-empty bins in the creation of the `M` array.

    This can be remedied by setting `self.Nmin = 0`, though this is generally
    not recommended as bins with N < 1 are physically meaningless.
    '''

    def __init__(self, m_breaks, a_slopes, nbins, FeH, tout, Ndot, f_BH,
                 *args, **kwargs):

        self._fBH_target = np.atleast_1d(f_BH)

        if self._fBH_target.size != np.atleast_1d(tout).size:
            mssg = "`f_BH` must be same size as desired output times `tout`"
            raise ValueError(mssg)

        if np.any(self._fBH_target < 0):
            mssg = f"`f_BH` ({self._fBH_target}) must be greater than 0"
            raise ValueError(mssg)

        # leave BH_ret_dyn as default, will be ignored. BH_ret_int is fine
        super().__init__(m_breaks, a_slopes, nbins, FeH, tout, Ndot,
                         *args, **kwargs)

    def _dyn_eject_BH(self, Mr_BH, Nr_BH, Mtot, fBH_target):
        '''Determine and remove BHs, to represent dynamical ejections.

        Determines and removes an amount of BHs from the BH mass bins, designed
        to reflect the effects of dynamical ejections of centrally concentrated
        BHs from a cluster.
        Proceeds from the heaviest BH bins to the lightest, removing all mass
        in each bin until the desired total BH mass fraction is reached.

        The target mass fraction *must* be less than the current
        (`Mr_BH.sum() / Mtot`) mass fraction, as mass will only be removed.

        Both input arrays are modified *in place*, as well as returned.

        Parameters
        ----------
        Mr_BH : ndarray
            Array[nbin] of the total initial masses of black holes in each
            BH mass bin.

        Nr_BH : ndarray
            Array[nbin] of the total initial numbers of black holes in each
            BH mass bin.

        Mtot : float
            The total cluster mass, before ejections. Needed to compute the
            current f_BH as BHs are removed.

        fBH_target : float
            The target BH mass fraction.

        Returns
        -------
        Mr_BH : ndarray
            Array[nbin] of the total final masses of black holes in each
            BH mass bin, after dynamical ejections

        Nr_BH : ndarray
            Array[nbin] of the total final numbers of black holes in each
            BH mass bin, after dynamical ejections
        '''

        def Mrem(Δfbh, Mb, Mt):
            '''Compute the amount of BH mass to remove required to reach Δfbh'''
            return (Mt**2 * Δfbh) / ((Mt * (1 + Δfbh)) - Mb)

        # Remove BH starting from Heavy to Light
        j = Mr_BH.size

        MBH = Mr_BH.sum()

        while (fBH_target < (fBH_current := MBH / Mtot)) and (j >= 0):
            j -= 1

            # Skip empty bins
            if Nr_BH[j] < self.Nmin:
                continue

            # Removed entirety of this bin
            if ((MBH - Mr_BH[j]) / (Mtot - Mr_BH[j])) >= fBH_target:

                MBH -= Mr_BH[j]
                Mtot -= Mr_BH[j]

                Mr_BH[j] = 0
                Nr_BH[j] = 0

                continue

            # Remove required fraction of the last affected bin
            else:

                Δfreq = fBH_current - fBH_target
                Mreq = Mrem(Δfreq, MBH, Mtot)

                mr_BH_j = Mr_BH[j] / Nr_BH[j]

                Mr_BH[j] -= Mreq
                Nr_BH[j] -= Mreq / (mr_BH_j)

                break

        return Mr_BH, Nr_BH

    def _evolve(self):
        '''Main population evolution function, to be called on init'''

        # ------------------------------------------------------------------
        # Initialize output arrays
        # ------------------------------------------------------------------

        self.nout = len(self.tout)

        # Stars

        self.Ns, self.alpha, self.Nr, self.Mr = self.massbins.blanks(
            'empty', extra_dims=[self.nout], packed=False, grouped_rem=True
        )

        self.Ms, self.ms = np.empty_like(self.Ns), np.empty_like(self.Ns)
        self.mr = self.Nr._make(np.empty_like(x) for x in self.Nr)

        self.rem_types = np.repeat(self.massbins.nbin._fields[1:],
                                   self.massbins.nbin[1:])

        # ------------------------------------------------------------------
        # Initialise ODE solver
        # ------------------------------------------------------------------

        t0 = 0.0
        y0 = self.massbins.initial_values()

        sol = ode(self._derivs)
        sol.set_integrator("dopri5", max_step=1e12, atol=1e-5, rtol=1e-5)
        sol.set_initial_value(y0, t=t0)

        # ------------------------------------------------------------------
        # Evolve
        # ------------------------------------------------------------------

        for ti in self.t:

            # --------------------------------------------------------------
            # Integrate ODE solver
            # --------------------------------------------------------------

            sol.integrate(ti)

            # --------------------------------------------------------------
            # if this time is in the desired output times extract solutions
            # --------------------------------------------------------------

            if ti in self.tout:

                iout = np.where(self.tout == ti)[0][0]

                # ----------------------------------------------------------
                # Extract the N, M and alphas for stars and remnants
                # ----------------------------------------------------------

                Ns, alpha, Nr, Mr = self.massbins.unpack_values(
                    sol.y, grouped_rem=True
                )

                bins_MS = self.massbins.turned_off_bins(self.compute_mto(ti))

                As = Ns / Pk(alpha, 1, *bins_MS)
                Ms = As * Pk(alpha, 2, *bins_MS)

                # ----------------------------------------------------------
                # Eject BHs, first through natal kicks, then dynamically
                # ----------------------------------------------------------

                # Check if any BH have been created
                if ti > self.compute_tms(self.IFMR.BH_mi.upper):

                    fBH_target = self._fBH_target[iout]

                    # First remove mass from all bins by natal kicks, if desired
                    #  Do it first because we dont control the exact amount so
                    #  this could make the target fBH invalid afterwards
                    if self.natal_kicks:
                        self._natal_kick_BH(Mr.BH, Nr.BH)  # do it all in-place

                    Mtot = np.r_[Mr].sum() + Ms.sum()
                    Mbhtot = Mr.BH.sum()
                    fBH_current = Mbhtot / Mtot

                    # can only make fBH go down by removing BHs
                    if fBH_target > fBH_current:
                        mssg = (
                            f"Target `f_BH` ({fBH_target}) is greater than "
                            f"f_BH formed at t={ti:.1f} ({fBH_current:.3f}; "
                            f"after natal kicks). Reduce `f_BH`, alter IMF "
                            f"slopes or turn off natal kicks."
                        )
                        raise ValueError(mssg)

                    # Remove dynamical BH ejections
                    #   (note different signature to EvolvedMF)
                    self._dyn_eject_BH(Mr.BH, Nr.BH, Mtot=Mtot,
                                       fBH_target=fBH_target)

                    # Reset, in case someone's curious
                    self.BH_ret_dyn = Mr.BH.sum() / Mbhtot

                # ----------------------------------------------------------
                # save values into output arrays
                # ----------------------------------------------------------

                self.alpha[iout, :] = alpha

                # Stars
                self.Ns[iout, :] = Ns
                self.Ms[iout, :] = Ms

                self.ms[iout, :] = Ms / Ns

                # Remnants
                for c in range(len(Nr)):  # Each remnant class

                    self.Nr[c][iout, :] = Nr[c]
                    self.Mr[c][iout, :] = Mr[c]

                    # Precise mr only matters when Nr > 0
                    mr = 0.5 * np.sum(self.massbins.bins[c + 1], axis=0)

                    rem_mask = Nr[c] > 0
                    mr[rem_mask] = Mr[c][rem_mask] / Nr[c][rem_mask]

                    self.mr[c][iout, :] = mr


class InitialBHPopulation:
    '''Generate an initial population of black holes from a mass function.

    Evolves and generates, from either a stellar IMF or an explicit BH mass
    function parametrization, the initial population of BHs which will form.
    This is similar, in general, to evolving a mass function as normal only
    up to the short age at which all of the most massive stars (which will
    make all of the BHs in a population) turn off and form remnants.

    As this class is *only* focused on the initial BH populations, the
    constituent DEs can be simplified, ignoring all other remnants (which will
    not form in time) and all stellar escapes (which are largely negligible
    by this time). This speeds up the computation considerably.

    This class should be initialized using one of the given classmethods,
    `from_IMF` (to evolve the cluster from a given IMF until all BHs are formed)
    or `from_BHMF` (to explicitly generate a population of BHs based on a given
    power-law BH mass function, without actually evolving anything).
    See the documentation on these methods for details on the available
    input parameters.

    Attributes
    ----------
    M, N, m : np.ndarray
        The total mass, total number and mean mass of BHs, in each mass bin.

    Mtot, Ntot : float
        The total mass and amount of BHs in the entire population.

    bins : mbin
        The BH mass bin edges.

    age : float
        The final age, in Myr, of the evolution, i.e. when all BHs are formed.
        Only available if using the `from_IMF` init method.

    Ms_lost, Ns_lost : float
        The total mass and amount of stars lost, from the initial population
        based on the IMF, in the formation of the BHs.
        Only available if using the `from_IMF` init method.
    '''

    # TODO add unit tests

    _natal_kick_BH = EvolvedMF._natal_kick_BH
    _get_retention_frac = EvolvedMF._get_retention_frac

    age = None
    Ms_lost = None
    Ns_lost = None

    def f_BH(self, M_cluster):
        '''The total mass fraction in BHs, given a total population mass.'''
        return self.Mtot / M_cluster

    def __init__(self, M_BH, N_BH, BH_bins, FeH, *, vesc=90., natal_kicks=True):
        '''Should not init from this, use provided classmethods.'''

        # ------------------------------------------------------------------
        # Attributes needed for stolen natal kick functions from `EvolvedMF`
        # ------------------------------------------------------------------

        self.Nmin = 0.1
        self.vesc = vesc

        # ------------------------------------------------------------------
        # Optionally perform all natal kicks on the input mass arrays
        # ------------------------------------------------------------------

        if natal_kicks:
            # load in the ifmr data to interpolate fb from mr
            feh_path = get_data(f"sse/MP_FEH{IFMR(FeH).FeH_BH:+.2f}.dat")
            self.fb_grid = np.loadtxt(feh_path, usecols=(1, 3), unpack=True)

            # Do the natal kicks, in place
            self._natal_kick_BH(M_BH, N_BH)

        # ------------------------------------------------------------------
        # Compute and store all final BH mass and number arrays and values
        # ------------------------------------------------------------------

        self.M = M_BH
        self.N = N_BH

        self.bins = BH_bins

        # get BH_mr
        zmsk = self.N > 0  # Precise mr only matters when Nr > 0
        self.m = 0.5 * np.sum(BH_bins, axis=0)
        self.m[zmsk] = self.M[zmsk] / self.N[zmsk]

        self.Mtot = self.M.sum()
        self.Ntot = self.N.sum()

    # ----------------------------------------------------------------------
    # Initializations
    # ----------------------------------------------------------------------

    @classmethod
    def from_IMF(cls, m_breaks, a_slopes, nbins, FeH, N0=5e5, *,
                 natal_kicks=True, vesc=90., binning_method='default'):
        '''Initialize a BH population by evolving from an IMF.

        Based on a given IMF, sharing many arguments with `EvolvedMF`, generate
        an initial stellar population and evolve it using simplified DEs to
        the age at which all BHs will have formed.

        As it is evolved from a stellar population, also stored is the age
        at which all BHs are formed, and the amount of stars which are lost
        to the formation of BHs (but not tidal escapes from a cluster).

        Parameters
        ----------
        m_breaks : list of float
            IMF break-masses (including outer bounds; size N+1).

        a_slopes : list of float
            IMF slopes α (size N).

        nbins : list of int
            Number of mass bins in each regime of the IMF, as defined by
            m_breaks (size N).

        FeH : float
            Metallicity, in solar fraction [Fe/H].

        N0 : int, optional
            Total initial number of stars, over all bins. Defaults to 5e5 stars.

        natal_kicks : bool, optional
            Whether to account for natal kicks in the BH dynamical retention.
            Defaults to True (note this difference from `EvolvedMF`).

        vesc : float, optional
            Initial cluster escape velocity, in km/s, for use in the
            computation of BH natal kick effects. Defaults to 90 km/s.

        binning_method : {'default', 'split_log', 'split_linear'}, optional
            The spacing method to use when constructing the mass bins.
            See `masses.MassBins` for more information.
        '''

        def compute_tms(mi):
            a = tms_constants
            return a[0] * np.exp(a[1] * mi ** a[2])

        def compute_mto(t):
            a0, a1, a2 = tms_constants
            if t > a0:
                return (np.log(t / a0) / a1) ** (1 / a2)
            else:
                return np.inf

        def _derivs_BHs(t, y):
            '''Derivatives relevant to mass changes due to stellar evolution.
            This is a massively simplified case of the DEs from `EvolvedMF`,
            and is only valid up to the age all BHs are formed.
            `y` is [MS N array, BH N array, BH M array].
            '''

            # Setup derivative bins
            Ns = y[:nbin_MS]  # only Ns is relevant right now
            dNs = np.zeros(nbin_MS)
            dNr, dMr = np.zeros(nbin_BH), np.zeros(nbin_BH)

            frem = 1.0  # Retain all BHs

            # Apply only if this time is atleast later than the earliest tms
            if t > tms_u[-1]:

                # Find out which mass bin is the current turn-off bin
                isev = np.where(t > tms_u)[0][0]

                mto, m1 = compute_mto(t), massbins.bins.MS.lower[isev]

                # Avoid "hitting" the bin edge
                if mto > m1 and (Nj := Ns[isev]) > 0.1:

                    # The normalization constant
                    # TODO deal with the constant divide-by-zero warning here
                    Aj = Nj / Pk(a_slopes[-1], 1, m1, mto)

                    # Get the number of turn-off stars per unit of mass
                    dNdm = Aj * mto ** a_slopes[-1]

                else:
                    dNdm = 0
                    # TODO just break???

                # Compute the full dN/dt = dN/dm * dm/dt
                a = tms_constants
                dmdt = abs((1.0 / (a[1] * a[2] * t))
                           * (np.log(t / a[0]) / a[1]) ** (1 / a[2] - 1))

                dNdt = -dNdm * dmdt

                # Fill in star derivatives
                dNs[isev] = dNdt

                # Skip 0-mass remnants
                if t <= final_age and (m_rem := _ifmr.predict(mto)) > 0:

                    # Find bin based on lower bin edge (must be careful later)
                    irem = np.flatnonzero(massbins.bins.BH.lower <= m_rem)[-1]

                    # Fill in remnant derivatives
                    dNr[irem] = -dNdt * frem
                    dMr[irem] = -m_rem * dNdt * frem

            return np.r_[dNs, dNr, dMr]

        # ------------------------------------------------------------------
        # Setup mass bins and initial values, based on the IMF
        # ------------------------------------------------------------------

        _ifmr = IFMR(FeH)

        massbins = MassBins(m_breaks, a_slopes, nbins, N0, _ifmr,
                            binning_method=binning_method)

        nbin_MS, nbin_BH = massbins.nbin.MS, massbins.nbin.BH

        # Load a_i coefficients derived from interpolated Dartmouth models
        mstogrid = np.loadtxt(get_data("sevtables/msto.dat"))
        nearest_FeH = np.argmin(np.abs(mstogrid[:, 0] - FeH))
        tms_constants = mstogrid[nearest_FeH, 1:]

        # Compute t_ms for all bin edges
        tms_u = compute_tms(massbins.bins.MS.upper)

        init_N, init_M = massbins.initial_values(packed=False)

        # ------------------------------------------------------------------
        # Integrate and solve the derivatives (evolve)
        # ------------------------------------------------------------------

        y0 = np.r_[init_N.MS, init_N.BH, init_M.BH]

        sol = ode(_derivs_BHs)
        sol.set_integrator("dopri5", max_step=1e12, atol=1e-5, rtol=1e-5)
        sol.set_initial_value(y0, t=0.0)

        final_age = compute_tms(_ifmr.BH_mi.lower + 0.1)  # fltpnt bound errors

        # Integrate the solver at each bin bound, to match EvolvedMF.
        # The ODE solvers are suspiciously sensitive to this, and while this
        # may not be the most sane solution, this will recreate EvolvedMF.
        for ti in np.sort(tms_u[tms_u < final_age]):
            sol.integrate(ti)

        sol.integrate(final_age)

        N_BH, M_BH = sol.y[nbin_MS:nbin_MS + nbin_BH], sol.y[nbin_MS + nbin_BH:]

        # ------------------------------------------------------------------
        # Create the final class instance, and populate some extra values
        # specific to this init method, related to the evolution
        # ------------------------------------------------------------------

        out = cls(M_BH, N_BH, massbins.bins.BH,
                  FeH=FeH, natal_kicks=natal_kicks, vesc=vesc)

        out.age = final_age

        Ns = sol.y[:nbin_MS]

        alphas = np.repeat(massbins.a, massbins._nbin_MS_each)
        As = Ns / Pk(alphas, 1, *massbins.bins.MS)
        Ms = As * Pk(alphas, 2, *massbins.bins.MS)

        # Stellar losses
        out.Ns_lost = init_N.MS.sum() - Ns.sum()
        out.Ms_lost = init_M.MS.sum() - Ms.sum()

        return out

    @classmethod
    def from_BHMF(cls, m_breaks, a_slopes, nbins, FeH, N0=1000, *,
                  natal_kicks=True, vesc=90.):
        '''Initialize a BH population based on a given power-law mass function.

        Based on an explicit given power-law parametrization of a BH mass
        function, generate a corresponding population of BH mass bins.
        This class *does not* simulate any evolution, and is not based on
        any sort of stellar mass function, but simply creates bins of mass
        (normalized to N0) for a given BH MF, and optionally provides
        natal kicks.

        Parameters
        ----------
        m_breaks : list of float
            IMF break-masses (including outer bounds; size 2 or 3).

        a_slopes : list of float
            IMF slopes α. Supports either a single or broken (i.e. double)
            power law (size 1 or 2).

        nbins : int or list of int
            Number of mass bins in each regime of the given MF, as defined by
            a_slopes/m_breaks.

        FeH : float
            Metallicity, in solar fraction [Fe/H].

        N0 : int, optional
            Total initial number of BHs, over all bins. Defaults to 1000 BHs.

        natal_kicks : bool, optional
            Whether to account for natal kicks in the BH dynamical retention.
            Defaults to True (note this difference from `EvolvedMF`).

        vesc : float, optional
            Initial cluster escape velocity, in km/s, for use in the
            computation of BH natal kick effects. Defaults to 90 km/s.

        binning_method : {'default', 'split_log', 'split_linear'}, optional
            The spacing method to use when constructing the mass bins.
            See `masses.MassBins` for more information.
        '''

        a, mb = a_slopes, m_breaks

        # ------------------------------------------------------------------
        # Compute the BH bin boundaries, based on the degree of the power law
        # ------------------------------------------------------------------

        if len(a) == 2:

            if isinstance(nbins, int):
                from .masses import _divide_bin_sizes
                nbins = _divide_bin_sizes(nbins, 2)

            A2 = (
                Pk(a[1], 1, mb[1], mb[2])
                + (mb[1] ** (a[1] - a[0]) * Pk(a[0], 1, mb[0], mb[1]))
            )**(-1)

            A1 = A2 * mb[1]**(a[1] - a[0])

            A = N0 * np.repeat([A1, A2], nbins)

            bin_sides = np.r_[
                np.geomspace(mb[0], mb[1], nbins[0] + 1),
                np.geomspace(mb[1], mb[2], nbins[1] + 1)[1:]
            ]

        elif len(a) == 1:

            A1 = Pk(a[0], 1, mb[0], mb[1])**(-1)

            A = N0 * np.repeat([A1], nbins)

            bin_sides = np.geomspace(mb[0], mb[1], nbins + 1)

        else:
            # TODO there's no real reason to restrict this honestly
            mssg = "`from_BHMF` only supports one or two component power laws"
            raise ValueError(mssg)

        bins = mbin(bin_sides[:-1], bin_sides[1:])

        # ------------------------------------------------------------------
        # Compute the number and mass of BHs in each bin
        # ------------------------------------------------------------------

        # Expand array of IMF slopes to all mass bins
        alpha = np.repeat(a, nbins)

        # Set initial star bins based on IMF
        N_BH = A * Pk(alpha, 1, *bins)
        M_BH = A * Pk(alpha, 2, *bins)

        return cls(M_BH, N_BH, bins, FeH=FeH,
                   natal_kicks=natal_kicks, vesc=vesc)
